# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Support for keyword-only fields in dataclasses for Python versions <3.10.

This module provides wrappers for `dataclasses.dataclass` and
`dataclasses.field` that simulate support for keyword-only fields for Python
versions before 3.10 (which is the version where dataclasses added keyword-only
field support).  If this module is imported in Python 3.10+, then
`kw_only_dataclasses.dataclass` and `kw_only_dataclasses.field` will simply be
aliases for `dataclasses.dataclass` and `dataclasses.field`.

For earlier Python versions, when constructing a dataclass, any fields that have
been marked as keyword-only (including inherited fields) will be moved to the
end of the constuctor's argument list. This makes it possible to have a base
class that defines a field with a default, and a subclass that defines a field
without a default. E.g.:

>>> from flax.linen import kw_only_dataclasses
>>> @kw_only_dataclasses.dataclass
... class Parent:
...   name: str = kw_only_dataclasses.field(default='', kw_only=True)

>>> @kw_only_dataclasses.dataclass
... class Child(Parent):
...   size: float  # required.

>>> import inspect
>>> print(inspect.signature(Child.__init__))
(self, size: float, name: str = '') -> None


(If we used `dataclasses` rather than `kw_only_dataclasses` for the above
example, then it would have failed with TypeError "non-default argument
'size' follows default argument.")

WARNING: fields marked as keyword-only will not *actually* be turned into
keyword-only parameters in the constructor; they will only be moved to the
end of the parameter list (after all non-keyword-only parameters).
"""

import dataclasses
import functools
import inspect
from types import MappingProxyType
from typing import Any, TypeVar

import typing_extensions as tpe

import flax

M = TypeVar('M', bound='flax.linen.Module')
FieldName = str
Annotation = Any
Default = Any


class _KwOnlyType:
  """Metadata tag used to tag keyword-only fields."""

  def __repr__(self):
    return 'KW_ONLY'


KW_ONLY = _KwOnlyType()


def field(*, metadata=None, kw_only=dataclasses.MISSING, **kwargs):
  """Wrapper for dataclassess.field that adds support for kw_only fields.

  Args:
    metadata: A mapping or None, containing metadata for the field.
    kw_only: If true, the field will be moved to the end of `__init__`'s
      parameter list.
    **kwargs: Keyword arguments forwarded to `dataclasses.field`

  Returns:
    A `dataclasses.Field` object.
  """
  if kw_only is not dataclasses.MISSING and kw_only:
    if (
      kwargs.get('default', dataclasses.MISSING) is dataclasses.MISSING
      and kwargs.get('default_factory', dataclasses.MISSING)
      is dataclasses.MISSING
    ):
      raise ValueError('Keyword-only fields with no default are not supported.')
    if metadata is None:
      metadata = {}
    metadata[KW_ONLY] = True
  return dataclasses.field(metadata=metadata, **kwargs)


@tpe.dataclass_transform(field_specifiers=(field,))  # type: ignore[literal-required]
def dataclass(cls=None, extra_fields=None, **kwargs):
  """Wrapper for dataclasses.dataclass that adds support for kw_only fields.

  Args:
    cls: The class to transform (or none to return a decorator).
    extra_fields: A list of `(name, type, Field)` tuples describing extra fields
      that should be added to the dataclass.  This is necessary for linen's
      use-case of this module, since the base class (linen.Module) is *not* a
      dataclass.  In particular, linen.Module class is used as the base for both
      frozen and non-frozen dataclass subclasses; but the frozen status of a
      dataclass must match the frozen status of any base dataclasses.
    **kwargs: Additional arguments for `dataclasses.dataclass`.

  Returns:
    `cls`.
  """

  def wrap(cls):
    return _process_class(cls, extra_fields=extra_fields, **kwargs)

  return wrap if cls is None else wrap(cls)


def _process_class(cls: type[M], extra_fields=None, **kwargs):
  """Transforms `cls` into a dataclass that supports kw_only fields."""
  if '__annotations__' not in cls.__dict__:
    cls.__annotations__ = {}

  # The original __dataclass_fields__ dicts for all base classes.  We will
  # modify these in-place before turning `cls` into a dataclass, and then
  # restore them to their original values.
  base_dataclass_fields = {}  # dict[cls, cls.__dataclass_fields__.copy()]

  # The keyword only fields from `cls` or any of its base classes.
  kw_only_fields: dict[FieldName, tuple[Annotation, Default]] = {}

  # Scan for KW_ONLY marker.
  kw_only_name = None
  for name, annotation in cls.__annotations__.items():
    if annotation is KW_ONLY:
      if kw_only_name is not None:
        raise TypeError('Multiple KW_ONLY markers')
      kw_only_name = name
    elif kw_only_name is not None:
      if not hasattr(cls, name):
        raise ValueError(
          'Keyword-only fields with no default are not supported.'
        )
      default = getattr(cls, name)
      if isinstance(default, dataclasses.Field):
        default.metadata = MappingProxyType({**default.metadata, KW_ONLY: True})
      else:
        default = field(default=default, kw_only=True)
      setattr(cls, name, default)
  if kw_only_name:
    del cls.__annotations__[kw_only_name]

  # Inject extra fields.
  if extra_fields:
    for name, annotation, default in extra_fields:
      if not (isinstance(name, str) and isinstance(default, dataclasses.Field)):
        raise ValueError(
          'Expected extra_fields to a be a list of '
          '(name, type, Field) tuples.'
        )
      setattr(cls, name, default)
      cls.__annotations__[name] = annotation

  # Extract kw_only fields from base classes' __dataclass_fields__.
  for base in reversed(cls.__mro__[1:]):
    if not dataclasses.is_dataclass(base):
      continue
    base_annotations = base.__dict__.get('__annotations__', {})
    base_dataclass_fields[base] = dict(
      getattr(base, '__dataclass_fields__', {})
    )
    for base_field in list(dataclasses.fields(base)):
      field_name = base_field.name
      if base_field.metadata.get(KW_ONLY) or field_name in kw_only_fields:
        kw_only_fields[field_name] = (
          base_annotations.get(field_name),
          base_field,
        )
        del base.__dataclass_fields__[field_name]

  # Remove any keyword-only fields from this class.
  cls_annotations = cls.__dict__['__annotations__']
  for name, annotation in list(cls_annotations.items()):
    value = getattr(cls, name, None)
    if (
      isinstance(value, dataclasses.Field) and value.metadata.get(KW_ONLY)
    ) or name in kw_only_fields:
      del cls_annotations[name]
      kw_only_fields[name] = (annotation, value)

  # Add keyword-only fields at the end of __annotations__, in the order they
  # were found in the base classes and in this class.
  for name, (annotation, default) in kw_only_fields.items():
    setattr(cls, name, default)
    cls_annotations.pop(name, None)
    cls_annotations[name] = annotation

  create_init = '__init__' not in vars(cls) and kwargs.get('init', True)

  # Apply the dataclass transform.
  transformed_cls: type[M] = dataclasses.dataclass(cls, **kwargs)

  # Restore the base classes' __dataclass_fields__.
  for _cls, fields in base_dataclass_fields.items():
    _cls.__dataclass_fields__ = fields

  if create_init:
    dataclass_init = transformed_cls.__init__
    # use sum to count the number of init fields that are not keyword-only
    expected_num_args = sum(
      f.init and not f.metadata.get(KW_ONLY, False)
      for f in dataclasses.fields(transformed_cls)
    )

    @functools.wraps(dataclass_init)
    def init_wrapper(self, *args, **kwargs):
      num_args = len(args)
      if num_args > expected_num_args:
        # we add + 1 to each to account for `self`, matching python's
        # default error message
        raise TypeError(
          f'__init__() takes {expected_num_args + 1} positional '
          f'arguments but {num_args + 1} were given'
        )

      dataclass_init(self, *args, **kwargs)

    init_wrapper.__signature__ = inspect.signature(dataclass_init)  # type: ignore
    transformed_cls.__init__ = init_wrapper  # type: ignore[method-assign]

  # Return the transformed dataclass
  return transformed_cls



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2021 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Hyperparameter configuration to run the example on TPUs."""

from configs import default as default_lib


def get_config():
  """Get the hyperparameter configuration to train on TPUs."""
  config = default_lib.get_config()

  # Consider setting the batch size to max(tpu_chips * 256, 8 * 1024) if you
  # train on a larger pod slice.
  config.batch_size = 1024
  config.shuffle_buffer_size = 16 * 1024
  config.cache = True
  config.half_precision = True

  return config


metrics = default_lib.metrics



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VAE model definitions."""

from flax import linen as nn
from jax import random
import jax.numpy as jnp


class Encoder(nn.Module):
  """VAE Encoder."""

  latents: int

  @nn.compact
  def __call__(self, x):
    x = nn.Dense(500, name='fc1')(x)
    x = nn.relu(x)
    mean_x = nn.Dense(self.latents, name='fc2_mean')(x)
    logvar_x = nn.Dense(self.latents, name='fc2_logvar')(x)
    return mean_x, logvar_x


class Decoder(nn.Module):
  """VAE Decoder."""

  @nn.compact
  def __call__(self, z):
    z = nn.Dense(500, name='fc1')(z)
    z = nn.relu(z)
    z = nn.Dense(784, name='fc2')(z)
    return z


class VAE(nn.Module):
  """Full VAE model."""

  latents: int = 20

  def setup(self):
    self.encoder = Encoder(self.latents)
    self.decoder = Decoder()

  def __call__(self, x, z_rng):
    mean, logvar = self.encoder(x)
    z = reparameterize(z_rng, mean, logvar)
    recon_x = self.decoder(z)
    return recon_x, mean, logvar

  def generate(self, z):
    return nn.sigmoid(self.decoder(z))


def reparameterize(rng, mean, logvar):
  std = jnp.exp(0.5 * logvar)
  eps = random.normal(rng, logvar.shape)
  return mean + eps * std


def model(latents):
  return VAE(latents=latents)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
import numpy as np
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax.core import Array, Scope, init, lift, nn, unfreeze

default_norm = partial(nn.batch_norm)


def residual_block(scope: Scope, x: Array, conv, norm, act, features: int):
  residual = x
  x = scope.child(conv, 'conv_1')(x, features, (3, 3))
  x = scope.child(norm, 'bn_1')(x)
  x = act(x)
  x = scope.child(conv, 'conv_2')(x, features, (3, 3))
  x = scope.child(norm, 'bn_2')(x)
  return act(residual + x)


def big_resnet(
  scope: Scope,
  x,
  blocks=(10, 5),
  dtype=jnp.float32,
  norm=default_norm,
  act=nn.relu,
):
  conv = partial(nn.conv, bias=False, dtype=dtype)
  norm = partial(norm, dtype=dtype)

  # a two stage resnet where inner blocks are rematerialized to make sure
  # memory consumtion grows as O(sqrt(N)) and compute is O(N) where N is the number of blocks..
  # we use a double scan such that the compiled binary is of size O(1).
  print('total residual blocks:', np.prod(blocks))

  def body_fn(scope, x):
    return residual_block(scope, x, conv, norm, act, features=x.shape[-1])

  return lift.remat_scan(
    body_fn,
    lengths=blocks,
    variable_axes={'params': 0, 'batch_stats': 0},
    split_rngs={'params': True},
    policy=None,
  )(scope, x)


class BigResnetTest(absltest.TestCase):
  def test_big_resnet(self):
    x = random.normal(random.key(0), (1, 8, 8, 8))
    y, variables = init(big_resnet)(random.key(1), x)
    self.assertEqual(y.shape, (1, 8, 8, 8))
    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    batch_stats_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['batch_stats'])
    )
    self.assertEqual(
      param_shapes,
      {
        'conv_1': {'kernel': (10, 5, 3, 3, 8, 8)},
        'conv_2': {'kernel': (10, 5, 3, 3, 8, 8)},
        'bn_1': {'scale': (10, 5, 8), 'bias': (10, 5, 8)},
        'bn_2': {'scale': (10, 5, 8), 'bias': (10, 5, 8)},
      },
    )
    self.assertEqual(
      batch_stats_shapes,
      {
        'bn_1': {'var': (10, 5, 8), 'mean': (10, 5, 8)},
        'bn_2': {'var': (10, 5, 8), 'mean': (10, 5, 8)},
      },
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.struct."""

import dataclasses
from typing import Any

import jax
from absl.testing import absltest, parameterized
from jax._src.tree_util import prefix_errors

from flax import struct

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


@struct.dataclass
class Point:
  x: float
  y: float
  meta: Any = struct.field(pytree_node=False)


class StructTest(parameterized.TestCase):
  def test_no_extra_fields(self):
    p = Point(x=1, y=2, meta={})
    with self.assertRaises(dataclasses.FrozenInstanceError):
      p.new_field = 1

  def test_mutation(self):
    p = Point(x=1, y=2, meta={})
    new_p = p.replace(x=3)
    self.assertEqual(new_p, Point(x=3, y=2, meta={}))
    with self.assertRaises(dataclasses.FrozenInstanceError):
      p.y = 3

  # Slots param is not supported in Python 3.9
  # TODO(lew): Uncomment when Flax upgrades to Python 3.10.

  # def test_slots(self):
  #   slots_dataclass = partial(struct.dataclass, frozen=False, slots=True)
  #   @slots_dataclass
  #   class SlotsPoint:
  #     x: float
  #     y: float
  #   p = SlotsPoint(x=1., y=2.)
  #   p.x = 3.  # can assign to existing fields
  #   self.assertEqual(p, SlotsPoint(x=3., y=2.))
  #   with self.assertRaises(AttributeError):
  #     p.z = 0.  # can't create new fields by accident.

  def test_pytree_nodes(self):
    p = Point(x=1, y=2, meta={'abc': True})
    leaves = jax.tree_util.tree_leaves(p)
    self.assertEqual(leaves, [1, 2])
    new_p = jax.tree_util.tree_map(lambda x: x + x, p)
    self.assertEqual(new_p, Point(x=2, y=4, meta={'abc': True}))

  def test_keypath_error(self):
    # TODO(mattjj): avoid using internal prefix_errors by testing vmap error msg
    (e,) = prefix_errors(Point(1.0, [2.0], meta={}), Point(1.0, 2.0, meta={}))
    with self.assertRaisesRegex(ValueError, r'in_axes\.y'):
      raise e('in_axes')

  def test_double_wrap_no_op(self):
    class A:
      a: int

    self.assertFalse(hasattr(A, '_flax_dataclass'))

    A = struct.dataclass(A)
    self.assertTrue(hasattr(A, '_flax_dataclass'))

    A = struct.dataclass(A)  # no-op
    self.assertTrue(hasattr(A, '_flax_dataclass'))

  def test_wrap_pytree_node_no_error(self):
    @struct.dataclass
    class A(struct.PyTreeNode):
      a: int

  # TODO(marcuschiam): Uncomment when Flax upgrades to Python 3.10.
  # @parameterized.parameters(
  #     {'mode': 'dataclass'},
  #     {'mode': 'pytreenode'},
  # )
  # def test_kw_only(self, mode):
  #   if mode == 'dataclass':
  #     @struct.dataclass
  #     class A:
  #       a: int = 1

  #     @functools.partial(struct.dataclass, kw_only=True)
  #     class B(A):
  #       b: int
  #   elif mode == 'pytreenode':
  #     class A(struct.PyTreeNode):
  #       a: int = 1

  #     class B(A, struct.PyTreeNode, kw_only=True):
  #       b: int

  #   obj = B(b=2)
  #   self.assertEqual(obj.a, 1)
  #   self.assertEqual(obj.b, 2)

  #   with self.assertRaisesRegex(TypeError, "non-default argument 'b' follows default argument"):
  #     if mode == 'dataclass':
  #       @struct.dataclass
  #       class B(A):
  #         b: int
  #     elif mode == 'pytreenode':
  #       class B(A, struct.PyTreeNode):
  #         b: int

  # TODO(marcuschiam): Uncomment when Flax upgrades to Python 3.10.
  # @parameterized.parameters(
  #     {'mode': 'dataclass'},
  #     {'mode': 'pytreenode'},
  # )
  # def test_mutable(self, mode):
  #   if mode == 'dataclass':
  #     @struct.dataclass
  #     class A:
  #       a: int = 1

  #     @functools.partial(struct.dataclass, frozen=False)
  #     class B:
  #       b: int = 1
  #   elif mode == 'pytreenode':
  #     class A(struct.PyTreeNode):
  #       a: int = 1

  #     class B(struct.PyTreeNode, frozen=False):
  #       b: int = 1

  #   obj = A()
  #   with self.assertRaisesRegex(dataclasses.FrozenInstanceError, "cannot assign to field 'a'"):
  #     obj.a = 2

  #   obj = B()
  #   obj.b = 2
  #   self.assertEqual(obj.b, 2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import optax

from absl.testing import absltest
import numpy as np

from flax import linen
from flax import nnx


class TrainState(nnx.TrainState):
  batch_stats: nnx.State


class TestHelpers:
  def test_train_state(self):
    m = nnx.Dict(a=nnx.Param(1), b=nnx.BatchStat(2))

    graphdef, params, batch_stats = nnx.split(m, nnx.Param, nnx.BatchStat)

    state = TrainState.create(
      graphdef,
      params=params,
      tx=optax.sgd(1.0),
      batch_stats=batch_stats,
    )

    leaves = jax.tree_util.tree_leaves(state)

  def test_train_state_methods(self):
    class Foo(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs):
        self.linear = nnx.Linear(2, 4, rngs=rngs)
        self.batch_norm = nnx.BatchNorm(4, rngs=rngs)

      def __call__(self, x: jax.Array, train: bool) -> jax.Array:
        x = self.linear(x)
        x = self.batch_norm(x, use_running_average=not train)
        return x

    module = Foo(rngs=nnx.Rngs(0))
    graphdef, params, batch_stats = nnx.split(module, nnx.Param, nnx.BatchStat)

    state = TrainState.create(
      graphdef,
      params=params,
      tx=optax.sgd(1.0),
      batch_stats=batch_stats,
    )

    x = jax.numpy.ones((1, 2))
    y, _updates = state.apply('params', 'batch_stats')(x, train=True)

    assert y.shape == (1, 4)

    # fake gradient
    grads = jax.tree_util.tree_map(jnp.ones_like, state.params)
    # test apply_gradients
    state = state.apply_gradients(grads)

  def test_nnx_linen_sequential_equivalence(self):
    key1, key2 = jax.random.split(jax.random.key(0), 2)
    rngs = nnx.Rngs(0)
    x = jax.random.uniform(key1, (3, 1, 5))

    model_nnx = nnx.Sequential(
      nnx.Linear(5, 4, rngs=rngs), nnx.Linear(4, 2, rngs=rngs)
    )
    model = linen.Sequential([linen.Dense(4), linen.Dense(2)])

    variables = model.init(key2, x)
    for layer_index in range(2):
      for param in ('kernel', 'bias'):
        variables['params'][f'layers_{layer_index}'][param] = getattr(
          model_nnx.layers[layer_index], param
        ).value
    out_nnx = model_nnx(x)
    out = model.apply(variables, x)
    np.testing.assert_array_equal(out, out_nnx)

    variables = model.init(key2, x)
    for layer_index in range(2):
      for param in ('kernel', 'bias'):
        getattr(model_nnx.layers[layer_index], param).value = variables[
          'params'
        ][f'layers_{layer_index}'][param]
    out_nnx = model_nnx(x)
    out = model.apply(variables, x)
    np.testing.assert_array_equal(out, out_nnx)


if __name__ == '__main__':
  absltest.main()




# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Callable

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax.core import Array, Scope, init, nn, unfreeze


def mlp(scope: Scope, x: Array, hidden: int, out: int):
  x = scope.child(nn.dense, 'hidden')(x, hidden)
  x = nn.relu(x)
  return scope.child(nn.dense, 'out')(x, out)


@dataclass
class AutoEncoder:
  latents: int
  features: int
  hidden: int

  def __call__(self, scope, x):
    z = self.encode(scope, x)
    return self.decode(scope, z)

  def encode(self, scope, x):
    return scope.child(mlp, 'encoder')(x, self.hidden, self.latents)

  def decode(self, scope, z):
    return scope.child(mlp, 'decoder')(z, self.hidden, self.features)


def module_method(fn, name=None):
  if name is None:
    name = fn.__name__ if hasattr(fn, '__name__') else None

  def wrapper(self, *args, **kwargs):
    scope = self.scope.rewound()
    mod_fn = lambda scope: fn(self, scope, *args, **kwargs)
    return scope.child(mod_fn, name)()

  return wrapper


@dataclass
class AutoEncoder2:
  scope: Scope
  latents: int
  features: int
  hidden: int

  def __call__(self, x):
    z = self.encode(x)
    return self.decode(z)

  @module_method
  def encode(self, scope, x):
    return mlp(scope, x, self.hidden, self.latents)

  @module_method
  def decode(self, scope, z):
    return mlp(scope, z, self.hidden, self.features)


@dataclass
class AutoEncoder3:
  encode: Callable
  decode: Callable

  @staticmethod
  def create(scope, hidden: int, latents: int, features: int):
    enc = scope.child(mlp, 'encode', hidden=hidden, out=latents)
    dec = scope.child(mlp, 'decode', hidden=hidden, out=features)
    return AutoEncoder3(enc, dec)

  def __call__(self, x):
    z = self.encode(x)
    return self.decode(z)


class AutoEncoderTest(absltest.TestCase):
  def test_auto_encoder_hp_struct(self):
    ae = AutoEncoder(latents=2, features=4, hidden=3)
    x = jnp.ones((1, 4))
    x_r, variables = init(ae)(random.key(0), x)
    self.assertEqual(x.shape, x_r.shape)
    variable_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(
      variable_shapes,
      {
        'encoder': {
          'hidden': {'kernel': (4, 3), 'bias': (3,)},
          'out': {'kernel': (3, 2), 'bias': (2,)},
        },
        'decoder': {
          'hidden': {'kernel': (2, 3), 'bias': (3,)},
          'out': {'kernel': (3, 4), 'bias': (4,)},
        },
      },
    )

  def test_auto_encoder_with_scope(self):
    ae = lambda scope, x: AutoEncoder2(scope, latents=2, features=4, hidden=3)(
      x
    )
    x = jnp.ones((1, 4))

    x_r, variables = init(ae)(random.key(0), x)
    self.assertEqual(x.shape, x_r.shape)
    variable_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(
      variable_shapes,
      {
        'encode': {
          'hidden': {'kernel': (4, 3), 'bias': (3,)},
          'out': {'kernel': (3, 2), 'bias': (2,)},
        },
        'decode': {
          'hidden': {'kernel': (2, 3), 'bias': (3,)},
          'out': {'kernel': (3, 4), 'bias': (4,)},
        },
      },
    )

  def test_auto_encoder_bind_method(self):
    ae = lambda scope, x: AutoEncoder3.create(
      scope, latents=2, features=4, hidden=3
    )(x)
    x = jnp.ones((1, 4))

    x_r, variables = init(ae)(random.key(0), x)
    self.assertEqual(x.shape, x_r.shape)
    variable_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(
      variable_shapes,
      {
        'encode': {
          'hidden': {'kernel': (4, 3), 'bias': (3,)},
          'out': {'kernel': (3, 2), 'bias': (2,)},
        },
        'decode': {
          'hidden': {'kernel': (2, 3), 'bias': (3,)},
          'out': {'kernel': (3, 4), 'bias': (4,)},
        },
      },
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark for the ImageNet example using fake data for quick perf results.

This script doesn't need the dataset, but it needs the dataset metadata.
That can be fetched with the script `flax/tests/download_dataset_metadata.sh`.
"""

import pathlib
import time

from absl.testing import absltest
from flax.testing import Benchmark
import jax
import tensorflow_datasets as tfds

# Local imports.
from configs import fake_data_benchmark as config_lib
import train

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class ImagenetBenchmarkFakeData(Benchmark):
  """Runs ImageNet using fake data for quickly measuring performance."""

  def test_fake_data(self):
    workdir = self.get_tmp_model_dir()
    config = config_lib.get_config()
    # Go two directories up to the root of the flax directory.
    flax_root_dir = pathlib.Path(__file__).absolute().parents[2]
    data_dir = str(flax_root_dir) + '/.tfds/metadata'

    # Warm-up first so that we are not measuring just compilation.
    with tfds.testing.mock_data(num_examples=1024, data_dir=data_dir):
      train.train_and_evaluate(config, workdir)

    start_time = time.time()
    with tfds.testing.mock_data(num_examples=1024, data_dir=data_dir):
      train.train_and_evaluate(config, workdir)
    benchmark_time = time.time() - start_time

    self.report_wall_time(benchmark_time)
    self.report_extras({
        'description': 'ImageNet ResNet50 with fake data',
        'model_name': 'resnet50',
        'parameters': f'hp=true,bs={config.batch_size}',
    })


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global configuration flags for Flax."""

import os
from contextlib import contextmanager
from typing import Any, Generic, NoReturn, TypeVar, overload

_T = TypeVar('_T')


class Config:
  # See https://google.github.io/pytype/faq.html.
  _HAS_DYNAMIC_ATTRIBUTES = True

  def __init__(self):
    self._values = {}

  def _add_option(self, name, default):
    if name in self._values:
      raise RuntimeError(f'Config option {name} already defined')
    self._values[name] = default

  def _read(self, name):
    try:
      return self._values[name]
    except KeyError:
      raise LookupError(f'Unrecognized config option: {name}')

  @overload
  def update(self, name: str, value: Any, /) -> None:
    ...

  @overload
  def update(self, holder: 'FlagHolder[_T]', value: _T, /) -> None:
    ...

  def update(self, name_or_holder, value, /):
    """Modify the value of a given flag.

    Args:
      name_or_holder: the name of the flag to modify or the corresponding
        flag holder object.
      value: new value to set.
    """
    name = name_or_holder
    if isinstance(name_or_holder, FlagHolder):
      name = name_or_holder.name
    if name not in self._values:
      raise LookupError(f'Unrecognized config option: {name}')
    self._values[name] = value


config = Config()

# Config parsing utils


class FlagHolder(Generic[_T]):
  def __init__(self, name, help):
    self.name = name
    self.__name__ = name[4:] if name.startswith('flax_') else name
    self.__doc__ = f'Flag holder for `{name}`.\n\n{help}'

  def __bool__(self) -> NoReturn:
    raise TypeError(
      "bool() not supported for instances of type '{0}' "
      "(did you mean to use '{0}.value' instead?)".format(type(self).__name__)
    )

  @property
  def value(self) -> _T:
    return config._read(self.name)


def bool_flag(name: str, *, default: bool, help: str) -> FlagHolder[bool]:
  """Set up a boolean flag.

  Example::

    enable_foo = bool_flag(
        name='flax_enable_foo',
        default=False,
        help='Enable foo.',
    )

  Now the ``FLAX_ENABLE_FOO`` shell environment variable can be used to
  control the process-level value of the flag, in addition to using e.g.
  ``config.update("flax_enable_foo", True)`` directly.

  Args:
    name: converted to lowercase to define the name of the flag. It is
      converted to uppercase to define the corresponding shell environment
      variable.
    default: a default value for the flag.
    help: used to populate the docstring of the returned flag holder object.

  Returns:
    A flag holder object for accessing the value of the flag.
  """
  name = name.lower()
  config._add_option(name, static_bool_env(name.upper(), default))
  fh = FlagHolder[bool](name, help)
  setattr(Config, name, property(lambda _: fh.value, doc=help))
  return fh


def static_bool_env(varname: str, default: bool) -> bool:
  """Read an environment variable and interpret it as a boolean.

  This is deprecated. Please use bool_flag() unless your flag
  will be used in a static method and does not require runtime updates.

  True values are (case insensitive): 'y', 'yes', 't', 'true', 'on', and '1';
  false values are 'n', 'no', 'f', 'false', 'off', and '0'.
  Args:
    varname: the name of the variable
    default: the default boolean value
  Returns:
    boolean return value derived from defaults and environment.
  Raises: ValueError if the environment variable is anything else.
  """
  val = os.getenv(varname, str(default))
  val = val.lower()
  if val in ('y', 'yes', 't', 'true', 'on', '1'):
    return True
  elif val in ('n', 'no', 'f', 'false', 'off', '0'):
    return False
  else:
    raise ValueError(
      'invalid truth value {!r} for environment {!r}'.format(val, varname)
    )


@contextmanager
def temp_flip_flag(var_name: str, var_value: bool):
  """Context manager to temporarily flip feature flags for test functions.

  Args:
    var_name: the config variable name (without the 'flax_' prefix)
    var_value: the boolean value to set var_name to temporarily
  """
  old_value = getattr(config, f'flax_{var_name}')
  try:
    config.update(f'flax_{var_name}', var_value)
    yield
  finally:
    config.update(f'flax_{var_name}', old_value)


# Flax Global Configuration Variables:

# Whether to use the lazy rng implementation.
flax_lazy_rng = static_bool_env('FLAX_LAZY_RNG', True)

flax_filter_frames = bool_flag(
  name='flax_filter_frames',
  default=True,
  help='Whether to hide flax-internal stack frames from tracebacks.',
)

flax_profile = bool_flag(
  name='flax_profile',
  default=True,
  help='Whether to run Module methods under jax.named_scope for profiles.',
)

flax_use_orbax_checkpointing = bool_flag(
  name='flax_use_orbax_checkpointing',
  default=True,
  help='Whether to use Orbax to save checkpoints.',
)

flax_preserve_adopted_names = bool_flag(
  name='flax_preserve_adopted_names',
  default=False,
  help="When adopting outside modules, don't clobber existing names.",
)

# TODO(marcuschiam): remove this feature flag once regular dict migration is complete
flax_return_frozendict = bool_flag(
  name='flax_return_frozendict',
  default=False,
  help='Whether to return FrozenDicts when calling init or apply.',
)

flax_fix_rng = bool_flag(
  name='flax_fix_rng_separator',
  default=False,
  help=(
    'Whether to add separator characters when folding in static data into'
    ' PRNG keys.'
  ),
)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2019 The SEED Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A class implementing minimal Atari 2600 preprocessing.
Adapted from SEED RL, originally adapted from Dopamine.
"""

import cv2
from gymnasium.spaces.box import Box
import gymnasium as gym
import numpy as np


class AtariPreprocessing:
  """A class implementing image preprocessing for Atari 2600 agents.
  Specifically, this provides the following subset from the JAIR paper
  (Bellemare et al., 2013) and Nature DQN paper (Mnih et al., 2015):
    * Frame skipping (defaults to 4).
    * Terminal signal when a life is lost (off by default).
    * Grayscale and max-pooling of the last two frames.
    * Downsample the screen to a square image (defaults to 84x84).
  More generally, this class follows the preprocessing guidelines set down in
  Machado et al. (2018), "Revisiting the Arcade Learning Environment:
  Evaluation Protocols and Open Problems for General Agents".
  It also provides random starting no-ops, which are used in the Rainbow, Apex
  and R2D2 papers.
  """

  def __init__(
      self,
      environment: gym.Env,
      frame_skip=4,
      terminal_on_life_loss=False,
      screen_size=84,
      max_random_noops=0,
  ):
    """Constructor for an Atari 2600 preprocessor.
    Args:
      environment: Gym environment whose observations are preprocessed.
      frame_skip: int, the frequency at which the agent experiences the game.
      terminal_on_life_loss: bool, If True, the step() method returns
        is_terminal=True whenever a life is lost. See Mnih et al. 2015.
      screen_size: int, size of a resized Atari 2600 frame.
      max_random_noops: int, maximum number of no-ops to apply at the beginning
        of each episode to reduce determinism. These no-ops are applied at a
        low-level, before frame skipping.
    Raises:
      ValueError: if frame_skip or screen_size are not strictly positive.
    """
    if frame_skip <= 0:
      raise ValueError(
          'Frame skip should be strictly positive, got {}'.format(frame_skip)
      )
    if screen_size <= 0:
      raise ValueError(
          'Target screen size should be strictly positive, got {}'.format(
              screen_size
          )
      )

    self.environment = environment
    self.terminal_on_life_loss = terminal_on_life_loss
    self.frame_skip = frame_skip
    self.screen_size = screen_size
    self.max_random_noops = max_random_noops

    obs_dims = self.environment.observation_space
    # Stores temporary observations used for pooling over two successive
    # frames.
    self.screen_buffer = [
        np.empty((obs_dims.shape[0], obs_dims.shape[1]), dtype=np.uint8),
        np.empty((obs_dims.shape[0], obs_dims.shape[1]), dtype=np.uint8),
    ]

    self.game_over = False
    self.lives = 0  # Will need to be set by reset().

  @property
  def observation_space(self):
    # Return the observation space adjusted to match the shape of the processed
    # observations.
    return Box(
        low=0,
        high=255,
        shape=(self.screen_size, self.screen_size, 1),
        dtype=np.uint8,
    )

  @property
  def action_space(self):
    return self.environment.action_space

  @property
  def reward_range(self):
    return self.environment.reward_range

  @property
  def metadata(self):
    return self.environment.metadata

  def close(self):
    return self.environment.close()

  def apply_random_noops(self):
    """Steps self.environment with random no-ops."""
    if self.max_random_noops <= 0:
      return
    # Other no-ops implementations actually always do at least 1 no-op. We
    # follow them.
    no_ops = self.environment.np_random.randint(1, self.max_random_noops + 1)
    for _ in range(no_ops):
      _, _, game_over, _, _ = self.environment.step(0)
      if game_over:
        self.environment.reset()

  def reset(self):
    """Resets the environment.
    Returns:
      observation: numpy array, the initial observation emitted by the
        environment.
    """
    self.environment.reset()
    self.apply_random_noops()

    self.lives = self.environment.unwrapped.ale.lives()  # pytype:disable=attribute-error
    self._fetch_grayscale_observation(self.screen_buffer[0])
    self.screen_buffer[1].fill(0)
    return self._pool_and_resize()

  def render(self, mode):
    """Renders the current screen, before preprocessing.
    This calls the Gym API's render() method.
    Args:
      mode: Mode argument for the environment's render() method.
        Valid values (str) are:
          'rgb_array': returns the raw ALE image.
          'human': renders to display via the Gym renderer.
    Returns:
      if mode='rgb_array': numpy array, the most recent screen.
      if mode='human': bool, whether the rendering was successful.
    """
    return self.environment.render(mode)  # pytype:disable=wrong-arg-count

  def step(self, action):
    """Applies the given action in the environment.
    Remarks:
      * If a terminal state (from life loss or episode end) is reached, this may
        execute fewer than self.frame_skip steps in the environment.
      * Furthermore, in this case the returned observation may not contain valid
        image data and should be ignored.
    Args:
      action: The action to be executed.
    Returns:
      observation: numpy array, the observation following the action.
      reward: float, the reward following the action.
      is_terminal: bool, whether the environment has reached a terminal state.
        This is true when a life is lost and terminal_on_life_loss, or when the
        episode is over.
      info: Gym API's info data structure.
    """
    accumulated_reward = 0.0

    for time_step in range(self.frame_skip):
      # We bypass the Gym observation altogether and directly fetch the
      # grayscale image from the ALE. This is a little faster.
      _, reward, game_over, _, info = self.environment.step(action)
      accumulated_reward += reward

      if self.terminal_on_life_loss:
        new_lives = self.environment.unwrapped.ale.lives()  # pytype:disable=attribute-error
        is_terminal = game_over or new_lives < self.lives
        self.lives = new_lives
      else:
        is_terminal = game_over

      if is_terminal:
        break
      # We max-pool over the last two frames, in grayscale.
      elif time_step >= self.frame_skip - 2:
        t = time_step - (self.frame_skip - 2)
        self._fetch_grayscale_observation(self.screen_buffer[t])

    # Pool the last two observations.
    observation = self._pool_and_resize()

    self.game_over = game_over
    return observation, accumulated_reward, is_terminal, info

  def _fetch_grayscale_observation(self, output):
    """Returns the current observation in grayscale.
    The returned observation is stored in 'output'.
    Args:
      output: numpy array, screen buffer to hold the returned observation.
    Returns:
      observation: numpy array, the current observation in grayscale.
    """
    self.environment.unwrapped.ale.getScreenGrayscale(output)  # pytype:disable=attribute-error
    return output

  def _pool_and_resize(self):
    """Transforms two frames into a Nature DQN observation.
    For efficiency, the transformation is done in-place in self.screen_buffer.
    Returns:
      transformed_screen: numpy array, pooled, resized screen.
    """
    # Pool if there are enough screens to do so.
    if self.frame_skip > 1:
      np.maximum(
          self.screen_buffer[0],
          self.screen_buffer[1],
          out=self.screen_buffer[0],
      )

    transformed_image = cv2.resize(
        self.screen_buffer[0],
        (self.screen_size, self.screen_size),
        interpolation=cv2.INTER_LINEAR,
    )
    int_image = np.asarray(transformed_image, dtype=np.uint8)
    return np.expand_dims(int_image, axis=2)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test policy by playing a full Atari game."""

import itertools
from typing import Any, Callable

import flax
import numpy as np

import agent
import env_utils


def policy_test(
    n_episodes: int,
    apply_fn: Callable[..., Any],
    params: flax.core.frozen_dict.FrozenDict,
    game: str,
):
  """Perform a test of the policy in Atari environment.

  Args:
    n_episodes: number of full Atari episodes to test on
    apply_fn: the actor-critic apply function
    params: actor-critic model parameters, they define the policy being tested
    game: defines the Atari game to test on

  Returns:
    total_reward: obtained score
  """
  test_env = env_utils.create_env(game, clip_rewards=False)
  for _ in range(n_episodes):
    obs = test_env.reset()
    state = obs[None, ...]  # add batch dimension
    total_reward = 0.0
    for t in itertools.count():
      log_probs, _ = agent.policy_action(apply_fn, params, state)
      probs = np.exp(np.array(log_probs, dtype=np.float32))
      probabilities = probs[0] / probs[0].sum()
      action = np.random.choice(probs.shape[1], p=probabilities)
      obs, reward, done, _ = test_env.step(action)
      total_reward += reward
      next_state = obs[None, ...] if not done else None
      state = next_state
      if done:
        break
  return total_reward



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.examples.mnist.mnist_lib."""

import pathlib
import tempfile

from absl.testing import absltest
import jax
from jax import numpy as jnp
import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from configs import default
import train


CNN_PARAMS = 825_034


class TrainTest(absltest.TestCase):
  """Test cases for train."""

  def setUp(self):
    super().setUp()
    # Make sure tf does not allocate gpu memory.
    tf.config.experimental.set_visible_devices([], "GPU")

  def test_cnn(self):
    """Tests CNN module used as the trainable model."""
    rng = jax.random.key(0)
    inputs = jnp.ones((1, 28, 28, 3), jnp.float32)
    output, variables = train.CNN().init_with_output(rng, inputs)

    self.assertEqual((1, 10), output.shape)
    self.assertEqual(
        CNN_PARAMS,
        sum(
            np.prod(arr.shape)
            for arr in jax.tree_util.tree_leaves(variables["params"])
        ),
    )

  def test_train_and_evaluate(self):
    """Tests training and evaluation code by running a single step."""
    # Create a temporary directory where tensorboard metrics are written.
    workdir = tempfile.mkdtemp()

    # Go two directories up to the root of the flax directory.
    flax_root_dir = pathlib.Path(__file__).parents[2]
    data_dir = str(flax_root_dir) + "/.tfds/metadata"  # pylint: disable=unused-variable

    # Define training configuration.
    config = default.get_config()
    config.num_epochs = 1
    config.batch_size = 8

    with tfds.testing.mock_data(num_examples=8, data_dir=data_dir):
      train.train_and_evaluate(config=config, workdir=workdir)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default Hyperparameter configuration."""

import ml_collections


def get_config():
  """Get the default hyperparameter configuration."""
  config = ml_collections.ConfigDict()

  # Path to load or store sentencepiece vocab file.
  config.vocab_path = None

  # Vocabulary size if `vocab_path` is not given.
  config.vocab_size = 32_000

  config.max_corpus_chars = 10**7

  # Name of TFDS translation dataset to use.
  config.dataset_name = 'wmt17_translate/de-en'

  # Optional name of TFDS translation dataset to use for evaluation.
  config.eval_dataset_name = 'wmt14_translate/de-en'
  config.eval_split = 'test'

  # Reverse the direction of translation.
  config.reverse_translation = False

  # Per device batch size for training.
  config.per_device_batch_size = 32

  # Beam size for inference.
  config.beam_size = 4

  config.num_train_steps = 100_000

  # Number of steps to take during evaluation.
  config.num_eval_steps = 20
  # Number of steps to generate predictions (used for BLEU score).
  # -1 will use the whole eval dataset.
  config.num_predict_steps = -1

  # Base learning rate.
  config.learning_rate = 0.002

  # Linear learning rate warmup.
  config.warmup_steps = 1000

  # Cross entropy loss label smoothing.
  config.label_smoothing = 0.1

  # Decay factor for AdamW style weight decay.
  config.weight_decay = 0.0

  # Maximum length cutoff for training examples.
  config.max_target_length = 256
  # Maximum length cutoff for eval examples.
  config.max_eval_target_length = 256
  # Maximum length cutoff for predicted tokens.
  config.max_predict_length = 256

  # Inputs and targets share embedding.
  config.share_embeddings = True

  # Final logit transform uses embedding matrix transpose.
  config.logits_via_embedding = True

  # Number of transformer layers.
  config.num_layers = 6

  # Size of query/key/value for attention.
  config.qkv_dim = 1024
  # Size of embeddings.
  config.emb_dim = 1024
  # Size of the MLP.
  config.mlp_dim = 4096

  # Number of attention heads.
  config.num_heads = 16

  # Dropout rate.
  config.dropout_rate = 0.1

  # Attention dropout rate.
  config.attention_dropout_rate = 0.1

  # Whether to save model checkpoints.
  config.save_checkpoints = True
  # Whether to restore from existing model checkpoints.
  config.restore_checkpoints = True

  # Save a checkpoint every these number of steps.
  config.checkpoint_every_steps = 10_000
  # Frequency of eval during training, e.g. every 1000 steps.
  config.eval_every_steps = 1_000

  # Use float16/bfloat16 (GPU/TPU) mixed precision training instead of float32.
  config.use_mixed_precision = True

  # Integer for PRNG random seed.
  config.seed = 0

  return config


def metrics():
  return [
      'train_loss',
      'eval_loss',
      'bleu',
      'eval_accuracy',
      'train_accuracy',
      'uptime',
      'steps_per_sec',
      'train_learning_rate',
  ]



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library file for executing training and evaluation on ogbg-molpcba."""

import os
from typing import Any, Dict, Iterable, Tuple, Optional

from absl import logging
from clu import checkpoint
from clu import metric_writers
from clu import metrics
from clu import parameter_overview
from clu import periodic_actions
import flax
import flax.core
import flax.linen as nn
from flax.training import train_state
import jax
import jax.numpy as jnp
import jraph
import ml_collections
import numpy as np
import optax
import sklearn.metrics
import tensorflow as tf

import input_pipeline
import models


def create_model(
    config: ml_collections.ConfigDict, deterministic: bool
) -> nn.Module:
  """Creates a Flax model, as specified by the config."""
  if config.model == 'GraphNet':
    return models.GraphNet(
        latent_size=config.latent_size,
        num_mlp_layers=config.num_mlp_layers,
        message_passing_steps=config.message_passing_steps,
        output_globals_size=config.num_classes,
        dropout_rate=config.dropout_rate,
        skip_connections=config.skip_connections,
        layer_norm=config.layer_norm,
        use_edge_model=config.use_edge_model,
        deterministic=deterministic,
    )
  if config.model == 'GraphConvNet':
    return models.GraphConvNet(
        latent_size=config.latent_size,
        num_mlp_layers=config.num_mlp_layers,
        message_passing_steps=config.message_passing_steps,
        output_globals_size=config.num_classes,
        dropout_rate=config.dropout_rate,
        skip_connections=config.skip_connections,
        layer_norm=config.layer_norm,
        deterministic=deterministic,
    )
  raise ValueError(f'Unsupported model: {config.model}.')


def create_optimizer(
    config: ml_collections.ConfigDict,
) -> optax.GradientTransformation:
  """Creates an optimizer, as specified by the config."""
  if config.optimizer == 'adam':
    return optax.adam(learning_rate=config.learning_rate)
  if config.optimizer == 'sgd':
    return optax.sgd(
        learning_rate=config.learning_rate, momentum=config.momentum
    )
  raise ValueError(f'Unsupported optimizer: {config.optimizer}.')


def binary_cross_entropy_with_mask(
    *, logits: jnp.ndarray, labels: jnp.ndarray, mask: jnp.ndarray
):
  """Binary cross entropy loss for unnormalized logits, with masked elements."""
  assert logits.shape == labels.shape == mask.shape
  assert len(logits.shape) == 2

  # To prevent propagation of NaNs during grad().
  # We mask over the loss for invalid targets later.
  labels = jnp.where(mask, labels, -1)

  # Numerically stable implementation of BCE loss.
  # This mimics TensorFlow's tf.nn.sigmoid_cross_entropy_with_logits().
  positive_logits = logits >= 0
  relu_logits = jnp.where(positive_logits, logits, 0)
  abs_logits = jnp.where(positive_logits, logits, -logits)
  return relu_logits - (logits * labels) + (jnp.log(1 + jnp.exp(-abs_logits)))


def predictions_match_labels(
    *, logits: jnp.ndarray, labels: jnp.ndarray, **kwargs
) -> jnp.ndarray:
  """Returns a binary array indicating where predictions match the labels."""
  del kwargs  # Unused.
  preds = logits > 0
  return (preds == labels).astype(jnp.float32)


def add_prefix_to_keys(result: Dict[str, Any], prefix: str) -> Dict[str, Any]:
  """Adds a prefix to the keys of a dict, returning a new dict."""
  return {f'{prefix}_{key}': val for key, val in result.items()}


@flax.struct.dataclass
class MeanAveragePrecision(
    metrics.CollectingMetric.from_outputs(('labels', 'logits', 'mask'))
):
  """Computes the mean average precision (mAP) over different tasks."""

  def compute(self):
    # Matches the official OGB evaluation scheme for mean average precision.
    values = super().compute()
    labels = values['labels']
    logits = values['logits']
    mask = values['mask']

    assert logits.shape == labels.shape == mask.shape
    assert len(logits.shape) == 2

    probs = jax.nn.sigmoid(logits)
    num_tasks = labels.shape[1]
    average_precisions = np.full(num_tasks, np.nan)

    for task in range(num_tasks):
      # AP is only defined when there is at least one negative data
      # and at least one positive data.
      is_labeled = mask[:, task]
      if len(np.unique(labels[is_labeled, task])) >= 2:
        average_precisions[task] = sklearn.metrics.average_precision_score(
            labels[is_labeled, task], probs[is_labeled, task]
        )

    # When all APs are NaNs, return NaN. This avoids raising a RuntimeWarning.
    if np.isnan(average_precisions).all():
      return np.nan
    return np.nanmean(average_precisions)


@flax.struct.dataclass
class EvalMetrics(metrics.Collection):
  accuracy: metrics.Average.from_fun(predictions_match_labels)
  loss: metrics.Average.from_output('loss')
  mean_average_precision: MeanAveragePrecision


@flax.struct.dataclass
class TrainMetrics(metrics.Collection):
  accuracy: metrics.Average.from_fun(predictions_match_labels)
  loss: metrics.Average.from_output('loss')


def replace_globals(graphs: jraph.GraphsTuple) -> jraph.GraphsTuple:
  """Replaces the globals attribute with a constant feature for each graph."""
  return graphs._replace(globals=jnp.ones([graphs.n_node.shape[0], 1]))


def get_predicted_logits(
    state: train_state.TrainState,
    graphs: jraph.GraphsTuple,
    rngs: Optional[Dict[str, jnp.ndarray]],
) -> jnp.ndarray:
  """Get predicted logits from the network for input graphs."""
  pred_graphs = state.apply_fn(state.params, graphs, rngs=rngs)
  logits = pred_graphs.globals
  return logits


def get_valid_mask(
    labels: jnp.ndarray, graphs: jraph.GraphsTuple
) -> jnp.ndarray:
  """Gets the binary mask indicating only valid labels and graphs."""
  # We have to ignore all NaN values - which indicate labels for which
  # the current graphs have no label.
  labels_mask = ~jnp.isnan(labels)

  # Since we have extra 'dummy' graphs in our batch due to padding, we want
  # to mask out any loss associated with the dummy graphs.
  # Since we padded with `pad_with_graphs` we can recover the mask by using
  # get_graph_padding_mask.
  graph_mask = jraph.get_graph_padding_mask(graphs)

  # Combine the mask over labels with the mask over graphs.
  return labels_mask & graph_mask[:, None]


@jax.jit
def train_step(
    state: train_state.TrainState,
    graphs: jraph.GraphsTuple,
    rngs: Dict[str, jnp.ndarray],
) -> Tuple[train_state.TrainState, metrics.Collection]:
  """Performs one update step over the current batch of graphs."""

  def loss_fn(params, graphs):
    curr_state = state.replace(params=params)

    # Extract labels.
    labels = graphs.globals

    # Replace the global feature for graph classification.
    graphs = replace_globals(graphs)

    # Compute logits and resulting loss.
    logits = get_predicted_logits(curr_state, graphs, rngs)
    mask = get_valid_mask(labels, graphs)
    loss = binary_cross_entropy_with_mask(
        logits=logits, labels=labels, mask=mask
    )
    mean_loss = jnp.sum(jnp.where(mask, loss, 0)) / jnp.sum(mask)

    return mean_loss, (loss, logits, labels, mask)

  grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  (_, (loss, logits, labels, mask)), grads = grad_fn(state.params, graphs)
  state = state.apply_gradients(grads=grads)

  metrics_update = TrainMetrics.single_from_model_output(
      loss=loss, logits=logits, labels=labels, mask=mask
  )
  return state, metrics_update


@jax.jit
def evaluate_step(
    state: train_state.TrainState,
    graphs: jraph.GraphsTuple,
) -> metrics.Collection:
  """Computes metrics over a set of graphs."""

  # The target labels our model has to predict.
  labels = graphs.globals

  # Replace the global feature for graph classification.
  graphs = replace_globals(graphs)

  # Get predicted logits, and corresponding probabilities.
  logits = get_predicted_logits(state, graphs, rngs=None)

  # Get the mask for valid labels and graphs.
  mask = get_valid_mask(labels, graphs)

  # Compute the various metrics.
  loss = binary_cross_entropy_with_mask(logits=logits, labels=labels, mask=mask)

  return EvalMetrics.single_from_model_output(
      loss=loss, logits=logits, labels=labels, mask=mask
  )


def evaluate_model(
    state: train_state.TrainState,
    datasets: Dict[str, tf.data.Dataset],
    splits: Iterable[str],
) -> Dict[str, metrics.Collection]:
  """Evaluates the model on metrics over the specified splits."""

  # Loop over each split independently.
  eval_metrics = {}
  for split in splits:
    split_metrics = None

    # Loop over graphs.
    for graphs in datasets[split].as_numpy_iterator():
      split_metrics_update = evaluate_step(state, graphs)

      # Update metrics.
      if split_metrics is None:
        split_metrics = split_metrics_update
      else:
        split_metrics = split_metrics.merge(split_metrics_update)
    eval_metrics[split] = split_metrics

  return eval_metrics  # pytype: disable=bad-return-type


def train_and_evaluate(
    config: ml_collections.ConfigDict, workdir: str
) -> train_state.TrainState:
  """Execute model training and evaluation loop.

  Args:
    config: Hyperparameter configuration for training and evaluation.
    workdir: Directory where the TensorBoard summaries are written to.

  Returns:
    The train state (which includes the `.params`).
  """
  # We only support single-host training.
  assert jax.process_count() == 1

  # Create writer for logs.
  writer = metric_writers.create_default_writer(workdir)
  writer.write_hparams(config.to_dict())

  # Get datasets, organized by split.
  logging.info('Obtaining datasets.')
  datasets = input_pipeline.get_datasets(
      config.batch_size,
      add_virtual_node=config.add_virtual_node,
      add_undirected_edges=config.add_undirected_edges,
      add_self_loops=config.add_self_loops,
  )
  train_iter = iter(datasets['train'])

  # Create and initialize the network.
  logging.info('Initializing network.')
  rng = jax.random.key(0)
  rng, init_rng = jax.random.split(rng)
  init_graphs = next(datasets['train'].as_numpy_iterator())
  init_graphs = replace_globals(init_graphs)
  init_net = create_model(config, deterministic=True)
  params = jax.jit(init_net.init)(init_rng, init_graphs)
  parameter_overview.log_parameter_overview(params)

  # Create the optimizer.
  tx = create_optimizer(config)

  # Create the training state.
  net = create_model(config, deterministic=False)
  state = train_state.TrainState.create(
      apply_fn=net.apply, params=params, tx=tx
  )

  # Set up checkpointing of the model.
  # The input pipeline cannot be checkpointed in its current form,
  # due to the use of stateful operations.
  checkpoint_dir = os.path.join(workdir, 'checkpoints')
  ckpt = checkpoint.Checkpoint(checkpoint_dir, max_to_keep=2)
  state = ckpt.restore_or_initialize(state)
  initial_step = int(state.step) + 1

  # Create the evaluation state, corresponding to a deterministic model.
  eval_net = create_model(config, deterministic=True)
  eval_state = state.replace(apply_fn=eval_net.apply)

  # Hooks called periodically during training.
  report_progress = periodic_actions.ReportProgress(
      num_train_steps=config.num_train_steps, writer=writer
  )
  profiler = periodic_actions.Profile(num_profile_steps=5, logdir=workdir)
  hooks = [report_progress, profiler]

  # Begin training loop.
  logging.info('Starting training.')
  train_metrics = None
  for step in range(initial_step, config.num_train_steps + 1):
    # Split PRNG key, to ensure different 'randomness' for every step.
    rng, dropout_rng = jax.random.split(rng)

    # Perform one step of training.
    with jax.profiler.StepTraceAnnotation('train', step_num=step):
      graphs = jax.tree_util.tree_map(np.asarray, next(train_iter))
      state, metrics_update = train_step(
          state, graphs, rngs={'dropout': dropout_rng}
      )

      # Update metrics.
      if train_metrics is None:
        train_metrics = metrics_update
      else:
        train_metrics = train_metrics.merge(metrics_update)

    # Quick indication that training is happening.
    logging.log_first_n(logging.INFO, 'Finished training step %d.', 10, step)
    for hook in hooks:
      hook(step)

    # Log, if required.
    is_last_step = step == config.num_train_steps - 1
    if step % config.log_every_steps == 0 or is_last_step:
      writer.write_scalars(
          step, add_prefix_to_keys(train_metrics.compute(), 'train')
      )
      train_metrics = None

    # Evaluate on validation and test splits, if required.
    if step % config.eval_every_steps == 0 or is_last_step:
      eval_state = eval_state.replace(params=state.params)

      splits = ['validation', 'test']
      with report_progress.timed('eval'):
        eval_metrics = evaluate_model(eval_state, datasets, splits=splits)
      for split in splits:
        writer.write_scalars(
            step, add_prefix_to_keys(eval_metrics[split].compute(), split)
        )

    # Checkpoint model, if required.
    if step % config.checkpoint_every_steps == 0 or is_last_step:
      with report_progress.timed('checkpoint'):
        ckpt.save(state)

  return state



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses

from absl.testing import absltest
import jax
import jax.numpy as jnp

from flax import nnx
from flax.nnx import compat


class TestCompatModule(absltest.TestCase):
  def test_compact_basic(self):
    class Linear(compat.Module):
      dout: int

      def setup(self):
        self.count = 0

      def __call__(self, x):
        self.count += 1
        if not hasattr(self, 'w'):
          assert self.scope is not None
          rngs = self.scope.rngs
          self.w = nnx.Param(
            jax.random.uniform(rngs(), (x.shape[-1], self.dout))
          )
          self.b = nnx.Param(jnp.zeros((self.dout,)))
        return x @ self.w + self.b[None]

    @dataclasses.dataclass
    class Foo(compat.Module):
      dout: int

      @compat.compact
      def __call__(self, x):
        din = x.shape[-1]
        self.linear = Linear(self.dout)
        x = self.linear(x)
        return x

    foo = Foo(5)
    x = jnp.ones((3, 2))
    rngs = nnx.Rngs(0)

    foo._set_scope(compat.Scope(rngs))
    y = foo(x)
    foo._set_scope(None)

    assert y.shape == (3, 5)
    assert hasattr(foo, 'Linear_0')

    assert foo.linear is foo.Linear_0
    assert foo.linear.count == 1
    assert rngs.default.count.value == 1

    foo._set_scope(compat.Scope(rngs))
    y = foo(x)
    foo._set_scope(None)

    assert foo.linear is foo.Linear_0
    assert foo.linear.count == 2

    # Rngs not called again
    assert rngs.default.count.value == 1

  def test_compact_parent_none(self):
    class Foo(compat.Module):
      pass

    class Bar(compat.Module):
      @compat.compact
      def __call__(self):
        return Foo().scope

    rngs = nnx.Rngs(0)
    bar = Bar()
    bar._set_scope(compat.Scope(rngs))
    scope = bar()
    bar._set_scope(None)
    assert bar.scope is None
    assert scope.rngs is rngs

    class Baz(compat.Module):
      @compat.compact
      def __call__(self):
        return Foo(parent=None).scope

    baz = Baz()
    baz._set_scope(compat.Scope(rngs))
    scope = baz()
    baz._set_scope(None)
    assert scope is None

  def test_name(self):
    class Foo(compat.Module):
      dout: int

      def __call__(self, x):
        if not hasattr(self, 'w'):
          assert self.scope is not None
          rngs = self.scope.rngs
          self.w = nnx.Param(
            jax.random.uniform(rngs(), (x.shape[-1], self.dout))
          )
        return x @ self.w

    class Bar(compat.Module):
      @compat.compact
      def __call__(self, x):
        return Foo(5, name='foo')(x)

    bar = Bar()
    x = jnp.ones((1, 2))
    rngs = nnx.Rngs(0)
    bar._set_scope(compat.Scope(rngs))
    y = bar(x)
    bar._set_scope(None)
    assert y.shape == (1, 5)

    assert hasattr(bar, 'foo')
    assert isinstance(bar.foo, Foo)

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Sequence

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax.core import Array, Scope, apply, init, lift, nn, unfreeze


def weight_std(fn, kernel_name='kernel', eps=1e-8):
  def std(variables):
    params = variables['params']
    assert kernel_name in params
    kernel = params[kernel_name]
    redux = tuple(range(kernel.ndim - 1))
    norm = jnp.square(kernel).sum(redux, keepdims=True)
    std_kernel = kernel / jnp.sqrt(norm + eps)
    params[kernel_name] = std_kernel
    return variables

  # map_variables handles a few of nasty edge cases here...
  # the transformed kind will be immutable inside fn
  # this way we avoid lost mutations to param
  # map_variables also avoids accidental reuse of rngs
  # and it makes sure that other state is updated correctly (not twice during init!)
  return lift.map_variables(fn, 'params', std, init=True)


def mlp(scope: Scope, x: Array, sizes: Sequence[int] = (8, 1)):
  std_dense = weight_std(
    partial(nn.dense, kernel_init=nn.initializers.normal(stddev=1e5))
  )
  for size in sizes[:-1]:
    x = scope.child(std_dense, prefix='hidden_')(x, size)
  return scope.child(nn.dense, 'out')(x, sizes[-1])


class WeightStdTest(absltest.TestCase):
  def test_weight_std(self):
    x = random.normal(
      random.key(0),
      (
        1,
        4,
      ),
    )
    y, variables = init(mlp)(random.key(1), x)

    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(
      param_shapes,
      {
        'hidden_0': {'kernel': (4, 8), 'bias': (8,)},
        'out': {'kernel': (8, 1), 'bias': (1,)},
      },
    )
    self.assertEqual(y.shape, (1, 1))
    self.assertTrue(y.ravel() < 1.0)

    y2 = apply(mlp)(variables, x)
    self.assertTrue(jnp.allclose(y, y2))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utils for positional embeddings (including RoPE)."""

import jax
import jax.numpy as jnp

_MAX_WAVELENGTH = 10_000

# This code uses implicit rank broadcast and needs this config to be 'allow',
# while the rest of Flax can use jax_numpy_rank_promotion=raise.
jax.config.update('jax_numpy_rank_promotion', 'allow')


def add_positional_embedding(
    input_embedding: jax.Array,
    position: int,
    max_wavelength: int = _MAX_WAVELENGTH,
) -> jax.Array:
  """Adds positional embeddings to input embeddings."""
  embed_dim = input_embedding.shape[-1]
  num_timescales = embed_dim // 2
  log_timescale_increment = jnp.log(float(max_wavelength)) / jnp.maximum(
      jnp.asarray(num_timescales, dtype=jnp.float32) - 1, 1
  )
  inv_timescales = jnp.exp(
      jnp.arange(num_timescales, dtype=jnp.float32) * -log_timescale_increment
  )
  scaled_time = position * inv_timescales
  signal = jnp.concatenate([jnp.sin(scaled_time), jnp.cos(scaled_time)])
  signal = jnp.pad(signal, [[0, jnp.mod(embed_dim, 2)]])
  position_embedding = signal.astype(jnp.float32)

  return input_embedding + position_embedding


def apply_rope(
    inputs: jax.Array,    # [B, L]
    positions: jax.Array, # [B, L]
    head_dim: int,
    max_wavelength: int = _MAX_WAVELENGTH,
) -> jax.Array:
  """Applies RoPE."""
  fraction = 2 * jnp.arange(0, head_dim // 2) / head_dim
  timescale = max_wavelength**fraction

  sinusoid_inp = (
      positions[..., jnp.newaxis] / timescale[jnp.newaxis, jnp.newaxis, :]
  )
  sinusoid_inp = sinusoid_inp[..., jnp.newaxis, :]
  sin = jnp.sin(sinusoid_inp)
  cos = jnp.cos(sinusoid_inp)

  first_half, second_half = jnp.split(inputs, 2, axis=-1)
  first_part = first_half * cos - second_half * sin
  second_part = second_half * cos + first_half * sin
  out = jnp.concatenate([first_part, second_part], axis=-1)
  return out.astype(inputs.dtype)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Base layers."""

from __future__ import annotations

from typing import Any, Sequence, Union

from flax import nnx
import flax.linen as nn
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike  # pylint: disable=g-importing-member,g-multiple-import


Shape = Sequence[Union[int, Any]]


class Einsum(nnx.Module):
  """Einsum is a convenience module for parameterized tensor multiplication."""

  def __init__(self, shape: Shape, *, rngs: nnx.Rngs):
    self.w = nnx.Param(nn.initializers.normal()(rngs.params(), shape))

  def __call__(self, eqn: str, x: ArrayLike) -> Array:
    return jnp.einsum(eqn, x, self.w.value)

  @property
  def shape(self) -> Shape:
    return self.w.value.shape


class RMSNorm(nnx.Module):
  """RMSNorm layer."""

  def __init__(self, dim: int, *, rngs: nnx.Rngs):
    self.scale = nnx.Param(nn.initializers.zeros_init()(rngs.params(), dim))

  def __call__(self, x):
    var = jnp.mean(jnp.square(x), axis=-1, keepdims=True)
    normed_inputs = jnp.asarray(x * jnp.reciprocal(jnp.sqrt(var + 1e-06)))
    # normed_inputs is a rank-K tensor, K > 1 (K is typically 2 or 3). scale is
    # a rank-1 tensor. To avoid implicit rank-promotion, reshape scale to
    # a (1, ..., 1, D) tensor, so the rank of scale matches normed_inputs.
    scale = jnp.expand_dims(self.scale.value, axis=range(len(x.shape) - 1))
    normed_inputs = normed_inputs * (1 + scale)
    return normed_inputs



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax Neural Network api."""

# pylint: disable=g-multiple-import
# re-export commonly used modules and functions
from flax.linen import activation as activation
from flax.linen import initializers as initializers
from flax.linen.activation import (
    celu as celu,
    elu as elu,
    gelu as gelu,
    glu as glu,
    leaky_relu as leaky_relu,
    log_sigmoid as log_sigmoid,
    log_softmax as log_softmax,
    relu as relu,
    sigmoid as sigmoid,
    silu as silu,
    soft_sign as soft_sign,
    softmax as softmax,
    softplus as softplus,
    swish as swish,
    tanh as tanh,
)
from flax.linen.pooling import (avg_pool as avg_pool, max_pool as max_pool)
from .attention import (
    dot_product_attention as dot_product_attention,
    multi_head_dot_product_attention as multi_head_dot_product_attention,
)
from .linear import (
    Embedding as Embedding,
    conv_transpose as conv_transpose,
    conv as conv,
    dense_general as dense_general,
    dense as dense,
    embedding as embedding,
)
from .normalization import (
    batch_norm as batch_norm,
    group_norm as group_norm,
    layer_norm as layer_norm,
)
from .stochastic import dropout as dropout

# pylint: enable=g-multiple-import



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import dataclasses
import threading
import typing as tp
from abc import ABC, abstractmethod

A = tp.TypeVar('A')
B = tp.TypeVar('B')


@dataclasses.dataclass
class ReprContext(threading.local):
  indent_stack: tp.List[str] = dataclasses.field(default_factory=lambda: [''])


REPR_CONTEXT = ReprContext()


@dataclasses.dataclass
class Object:
  type: tp.Union[str, type]
  start: str = '('
  end: str = ')'
  value_sep: str = '='
  elem_indent: str = '  '
  empty_repr: str = ''


@dataclasses.dataclass
class Attr:
  key: str
  value: tp.Union[str, tp.Any]
  start: str = ''
  end: str = ''


class Representable(ABC):
  __slots__ = ()

  @abstractmethod
  def __nnx_repr__(self) -> tp.Iterator[tp.Union[Object, Attr]]:
    raise NotImplementedError

  def __repr__(self) -> str:
    return get_repr(self)


@contextlib.contextmanager
def add_indent(indent: str) -> tp.Iterator[None]:
  REPR_CONTEXT.indent_stack.append(REPR_CONTEXT.indent_stack[-1] + indent)

  try:
    yield
  finally:
    REPR_CONTEXT.indent_stack.pop()


def get_indent() -> str:
  return REPR_CONTEXT.indent_stack[-1]


def get_repr(obj: Representable) -> str:
  if not isinstance(obj, Representable):
    raise TypeError(f'Object {obj!r} is not representable')

  iterator = obj.__nnx_repr__()
  config = next(iterator)
  if not isinstance(config, Object):
    raise TypeError(f'First item must be Config, got {type(config).__name__}')

  def _repr_elem(elem: tp.Any) -> str:
    if not isinstance(elem, Attr):
      raise TypeError(f'Item must be Elem, got {type(elem).__name__}')

    value = elem.value if isinstance(elem.value, str) else repr(elem.value)

    value = value.replace('\n', '\n' + config.elem_indent)

    return f'{config.elem_indent}{elem.start}{elem.key}{config.value_sep}{value}{elem.end}'

  with add_indent(config.elem_indent):
    elems = ',\n'.join(map(_repr_elem, iterator))

  if elems:
    elems = '\n' + elems + '\n'
  else:
    elems = config.empty_repr

  type_repr = (
    config.type if isinstance(config.type, str) else config.type.__name__
  )

  return f'{type_repr}{config.start}{elems}{config.end}'

class MappingReprMixin(tp.Mapping[A, B]):
  def __nnx_repr__(self):
    yield Object(type='', value_sep=': ', start='{', end='}')

    for key, value in self.items():
      yield Attr(repr(key), value)

@dataclasses.dataclass(repr=False)
class PrettyMapping(Representable):
  mapping: tp.Mapping

  def __nnx_repr__(self):
    yield Object(type='', value_sep=': ', start='{', end='}')

    for key, value in self.mapping.items():
      yield Attr(repr(key), value)


# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Language Modeling example.

This script trains a Transformer on a LM1B dataset.
"""

# pytype: disable=wrong-arg-count
# pytype: disable=attribute-error

import collections
import os

from absl import logging
from clu import metric_writers
from clu import periodic_actions
from flax import linen as nn
from flax.training import checkpoints
from flax.training import common_utils
import jax
from jax import random
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P, Mesh, NamedSharding
import ml_collections
import numpy as np
import optax
import tensorflow as tf

import input_pipeline
import models
import temperature_sampler
import utils


def rsqrt_schedule(
    init_value: float,
    shift: int = 0,
):
  """Applies a reverse square-root schedule.

  The reverse square root schedule is simply `lr = init_value / sqrt(step)`.

  Args:
    init_value: Base learning rate (before applying the rsqrt schedule).
    shift: How many steps the rsqrt should be shifted. Shifting the rsqrt
      schedule makes it less steep in the beginning (close to 0).

  Returns:
    A schedule that applies the reverse square root.
  """

  def schedule(count):
    return init_value * (count + shift) ** -0.5 * shift**0.5

  return schedule


def create_learning_rate_schedule(learning_rate: float, warmup_steps: int):
  """Creates a rsqrt schedule with linear warmup."""
  return optax.join_schedules(
      [
          optax.linear_schedule(
              init_value=0,
              end_value=learning_rate,
              transition_steps=warmup_steps,
          ),
          rsqrt_schedule(init_value=learning_rate, shift=warmup_steps),
      ],
      boundaries=[warmup_steps],
  )


def compute_weighted_cross_entropy(
    logits, targets, weights=None, label_smoothing=0.0
):
  """Compute weighted cross entropy and entropy for log probs and targets.

  Args:
   logits: [batch, length, num_classes] float array.
   targets: categorical targets [batch, length] int array.
   weights: None or array of shape [batch, length].
   label_smoothing: label smoothing constant, used to determine the on and off
     values.

  Returns:
    Tuple of scalar loss and batch normalizing factor.
  """
  if logits.ndim != targets.ndim + 1:
    raise ValueError(
        "Incorrect shapes. Got shape %s logits and %s targets"
        % (str(logits.shape), str(targets.shape))
    )
  vocab_size = logits.shape[-1]
  confidence = 1.0 - label_smoothing
  low_confidence = (1.0 - confidence) / (vocab_size - 1)
  normalizing_constant = -(
      confidence * jnp.log(confidence)
      + (vocab_size - 1) * low_confidence * jnp.log(low_confidence + 1e-20)
  )
  soft_targets = common_utils.onehot(
      targets, vocab_size, on_value=confidence, off_value=low_confidence
  )

  loss = -jnp.sum(soft_targets * nn.log_softmax(logits), axis=-1)
  loss = loss - normalizing_constant

  normalizing_factor = np.prod(targets.shape)
  if weights is not None:
    loss = loss * weights
    normalizing_factor = weights.sum()

  return loss.sum(), normalizing_factor


def compute_weighted_accuracy(logits, targets, weights=None):
  """Compute weighted accuracy for log probs and targets.

  Args:
   logits: [batch, length, num_classes] float array.
   targets: categorical targets [batch, length] int array.
   weights: None or array of shape [batch, length]

  Returns:
    Tuple of scalar loss and batch normalizing factor.
  """
  if logits.ndim != targets.ndim + 1:
    raise ValueError(
        "Incorrect shapes. Got shape %s logits and %s targets"
        % (str(logits.shape), str(targets.shape))
    )
  loss = jnp.equal(jnp.argmax(logits, axis=-1), targets)
  normalizing_factor = np.prod(logits.shape[:-1])
  if weights is not None:
    loss = loss * weights
    normalizing_factor = weights.sum()

  return loss.sum(), normalizing_factor


def compute_metrics(logits, labels, weights, label_smoothing=0.0):
  """Compute summary metrics."""
  loss, weight_sum = compute_weighted_cross_entropy(
      logits, labels, weights, label_smoothing
  )
  acc, _ = compute_weighted_accuracy(logits, labels, weights)
  metrics = {
      "loss": loss,
      "accuracy": acc,
      "denominator": weight_sum,
  }
  return metrics


# Primary training / eval / decode step functions.
# -----------------------------------------------------------------------------


def train_step(
    state,
    batch,
    config,
    learning_rate_fn,
    label_smoothing=0.0,
    dropout_rng=None,
):
  """Perform a single training step."""
  # X_position and X_segmentation are needed only when using "packed examples"
  # where multiple sequences are packed into the same example with this
  # metadata.
  # if such features are not present they are ignored and the example is treated
  # like a normal, unpacked sequence example.
  train_keys = ["inputs", "inputs_position", "inputs_segmentation"]
  (inputs, inputs_positions, inputs_segmentation) = (
      batch.get(k, None) for k in train_keys
  )

  weights = jnp.where(inputs > 0, 1, 0).astype(jnp.float32)

  dropout_rng = jax.random.fold_in(dropout_rng, state.step)

  def loss_fn(params):
    """loss function used for training."""
    logits = models.TransformerLM(config).apply(
        {"params": params},
        inputs,
        inputs_positions=inputs_positions,
        inputs_segmentation=inputs_segmentation,
        rngs={"dropout": dropout_rng},
    )

    loss, weight_sum = compute_weighted_cross_entropy(
        logits, inputs, weights, label_smoothing
    )
    mean_loss = loss / weight_sum
    return mean_loss, logits

  step = state.step
  lr = learning_rate_fn(step)
  grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  (_, logits), grads = grad_fn(state.params)
  new_state = state.apply_gradients(grads=grads)
  metrics = compute_metrics(logits, inputs, weights)
  metrics["learning_rate"] = lr

  return new_state, metrics


def eval_step(params, batch, config, label_smoothing=0.0):
  """Calculate evaluation metrics on a batch."""
  inputs = batch["inputs"]
  weights = jnp.where(inputs > 0, 1.0, 0.0)
  logits = models.TransformerLM(config).apply({"params": params}, inputs)

  return compute_metrics(logits, inputs, weights, label_smoothing)


def predict_step(
    inputs, params, rngkey, eos_id, max_decode_len, config, temperature, top_k
):
  """Predict language model on a batch."""
  target_shape = (inputs.shape[0], max_decode_len) + inputs.shape[2:]
  initial_variables = models.TransformerLM(config).init(
      jax.random.PRNGKey(0), jnp.ones(target_shape, config.dtype)
  )
  cache = initial_variables["cache"]

  def tokens_ids_to_logits(flat_ids, flat_cache):
    """Token slice to logits from decoder model."""
    # --> [batch * beam, 1, vocab]
    flat_logits, new_vars = models.TransformerLM(config).apply(
        {"params": params, "cache": flat_cache}, flat_ids, mutable=["cache"]
    )
    new_flat_cache = new_vars["cache"]
    # Remove singleton sequence-length dimension:
    # [batch, 1, vocab] --> [batch, vocab]
    flat_logits = flat_logits.squeeze(axis=1)
    return flat_logits, new_flat_cache

  # Using the above-defined single-step decoder function, run a
  # beam search over possible sequences given input encoding.
  seqs = temperature_sampler.temperature_sample(
      inputs,
      cache,
      tokens_ids_to_logits,
      rngkey,
      temperature=temperature,
      topk=top_k,
      eos_token=eos_id,
  )

  return seqs


# Utils for prediction
# -----------------------------------------------------------------------------


def pad_examples(x, desired_batch_size):
  """Expand batch to desired size by repeating last slice."""
  batch_pad = desired_batch_size - x.shape[0]
  return np.concatenate([x, np.tile(x[-1], (batch_pad, 1))], axis=0)


def per_host_sum_pmap(in_tree):
  """Execute psum on in_tree"s leaves over one device per host."""
  host2devices = collections.defaultdict(list)
  for d in jax.devices():
    host2devices[d.process_index].append(d)
  devices = [host2devices[k][0] for k in host2devices]
  host_psum = jax.pmap(lambda x: jax.lax.psum(x, "i"), "i", devices=devices)

  def pre_pmap(xs):
    return jax.tree_util.tree_map(
        lambda x: jnp.broadcast_to(x, (1,) + x.shape), xs
    )

  def post_pmap(xs):
    return jax.tree_util.tree_map(lambda x: x[0], xs)

  return post_pmap(host_psum(pre_pmap(in_tree)))


def tohost(x):
  """Collect batches from all devices to host and flatten batch dimensions."""
  n_device, n_batch, *remaining_dims = x.shape
  return np.array(x).reshape((n_device * n_batch,) + tuple(remaining_dims))


def evaluate(
    *,
    jit_eval_step,
    params,
    eval_ds: tf.data.Dataset,
    num_eval_steps: int,
    config,
):
  """Evaluate the target an return a dictionary with the metrics."""
  logging.info("Gathering evaluation metrics.")
  eval_metrics = []
  eval_iter = iter(eval_ds)  # pytype: disable=wrong-arg-types
  for _, eval_batch in zip(range(num_eval_steps), eval_iter):
    eval_batch = jax.tree_util.tree_map(lambda x: x._numpy(), eval_batch)  # pylint: disable=protected-access
    metrics = jit_eval_step(params, eval_batch, config)
    eval_metrics.append(metrics)
  eval_metrics = common_utils.stack_forest(eval_metrics)
  eval_metrics_sums = jax.tree_util.tree_map(jnp.sum, eval_metrics)
  eval_denominator = eval_metrics_sums.pop("denominator")
  eval_summary = jax.tree_util.tree_map(
      lambda x: x / eval_denominator,  # pylint: disable=cell-var-from-loop
      eval_metrics_sums,
  )
  return eval_summary


def generate_prediction(
    *,
    jit_pred_step,
    params,
    tokenized_prompts,
    eos_id,
    inference_rng,
    decode_tokens,
    config,
    predict_config,
):
  """Generate text from the prompt."""
  n_devices = jax.local_device_count()

  logging.info("Generating text.")
  predictions = []
  # Use batch of prompts provided by user.
  for pred_batch in jnp.array_split(
      tokenized_prompts, int(np.ceil(len(tokenized_prompts) / n_devices))
  ):
    cur_pred_batch_size = pred_batch.shape[0]
    if cur_pred_batch_size % n_devices:
      padded_size = int(np.ceil(cur_pred_batch_size / n_devices) * n_devices)
      pred_batch = jax.tree_util.tree_map(
          lambda x: pad_examples(x, padded_size), pred_batch
      )  # pylint: disable=cell-var-from-loop
    pred_batch = common_utils.shard(pred_batch)
    inference_rng, sub_rng = random.split(inference_rng)
    inference_rngs = random.split(sub_rng, n_devices)

    predicted = jit_pred_step(
        pred_batch,
        params,
        inference_rngs,
        eos_id,
        config.max_predict_length,
        predict_config,
        config.sampling_temperature,
        config.sampling_top_k,
    )
    predicted = tohost(predicted)
    # Iterate through non-padding examples of batch.
    for s in predicted[:cur_pred_batch_size]:
      prediction = decode_tokens(s)
      logging.info("Sample: %s", str(prediction))
      predictions.append(prediction)

    # Save generated texts for tensorboard.
    exemplars = ""
    for prediction in predictions:
      exemplars += f"{prediction}\n\n"
  return exemplars


def train_and_evaluate(config: ml_collections.ConfigDict, workdir: str):
  """Runs a training and evaluation loop.

  Args:
    config: Configuration to use.
    workdir: Working directory for checkpoints and TF summaries. If this
      contains checkpoint training will be resumed from the latest checkpoint.
  """
  tf.io.gfile.makedirs(workdir)

  vocab_path = config.vocab_path
  if vocab_path is None:
    vocab_path = os.path.join(workdir, "sentencepiece_model")
    config.vocab_path = vocab_path
  tf.io.gfile.makedirs(os.path.split(vocab_path)[0])

  # Load Dataset
  # ---------------------------------------------------------------------------
  logging.info("Initializing dataset.")
  train_ds, eval_ds, _, encoder = input_pipeline.get_datasets(
      n_devices=jax.local_device_count(), config=config, vocab_path=vocab_path
  )

  train_iter = iter(train_ds)
  vocab_size = int(encoder.vocab_size())
  eos_id = temperature_sampler.EOS_ID  # Default Sentencepiece EOS token.

  def decode_tokens(toks):
    valid_toks = toks[: np.argmax(toks == eos_id) + 1].astype(np.int32)
    return encoder.detokenize(valid_toks).numpy().decode("utf-8")

  def encode_strings(strs, max_len):
    tokenized_batch = np.zeros((len(strs), max_len), np.int32)
    for i, s in enumerate(strs):
      toks = encoder.tokenize(s).numpy()
      # Remove EOS token in prompt.
      tokenized_batch[i, : toks.shape[0] - 1] = toks[:-1]
    return tokenized_batch

  tokenized_prompts = encode_strings(
      [config.prompts], config.max_predict_length
  )

  logging.info("Initializing model, optimizer, and step functions.")
  # Build Model and Optimizer
  # ---------------------------------------------------------------------------
  train_config = models.TransformerConfig(
      vocab_size=vocab_size,
      output_vocab_size=vocab_size,
      logits_via_embedding=config.logits_via_embedding,
      dtype=jnp.bfloat16 if config.use_bfloat16 else jnp.float32,
      emb_dim=config.emb_dim,
      num_heads=config.num_heads,
      num_layers=config.num_layers,
      qkv_dim=config.qkv_dim,
      mlp_dim=config.mlp_dim,
      max_len=max(config.max_target_length, config.max_eval_target_length),
      dropout_rate=config.dropout_rate,
      attention_dropout_rate=config.attention_dropout_rate,
      deterministic=False,
      decode=False,
      kernel_init=nn.initializers.xavier_uniform(),
      bias_init=nn.initializers.normal(stddev=1e-6),
  )
  eval_config = train_config.replace(deterministic=True)
  predict_config = train_config.replace(deterministic=True, decode=True)

  # Mesh definition
  devices_array = utils.create_device_mesh(config)
  mesh = Mesh(devices_array, config.mesh_axes)

  start_step = 0
  rng = jax.random.PRNGKey(config.seed)
  rng, init_rng = jax.random.split(rng)
  rng, inference_rng = random.split(rng)

  m = models.TransformerLM(eval_config)

  learning_rate_fn = create_learning_rate_schedule(
      learning_rate=config.learning_rate, warmup_steps=config.warmup_steps
  )

  optimizer = optax.adamw(
      learning_rate_fn,
      b1=0.9,
      b2=0.98,
      eps=1e-9,
      weight_decay=config.weight_decay,
  )

  state, state_mesh_annotations = utils.setup_initial_state(
      m, optimizer, config, init_rng, mesh
  )
  data_sharding = NamedSharding(mesh, P(config.data_sharding))

  if config.restore_checkpoints:
    # Restore unreplicated optimizer + model state from last checkpoint.
    state = checkpoints.restore_checkpoint(workdir, state)
    # Grab last step.
    start_step = int(state.step)

  writer = metric_writers.create_default_writer(
      workdir, just_logging=jax.process_index() > 0
  )
  if start_step == 0:
    writer.write_hparams(dict(config))

  # compile multidevice versions of train/eval/predict step fn.
  jit_train_step = jax.jit(
      train_step,
      in_shardings=(
          state_mesh_annotations,
          data_sharding,
          None,
      ),  # type: ignore
      out_shardings=(state_mesh_annotations, None),  # type: ignore
      static_argnums=(2, 3, 4),
      donate_argnums=0,
  )

  jit_eval_step = jax.jit(
      eval_step,
      in_shardings=(
          state_mesh_annotations.params,
          data_sharding,
      ),  # type: ignore
      out_shardings=None,  # type: ignore
      static_argnums=(2, 3),
  )

  # Since the inputs and rngkey args for predict_step will be batched,
  # we must vmap them, otherwise the global arrays will be seen in each device
  jit_pred_step = jax.jit(
      jax.vmap(
          predict_step,
          in_axes=(
              0,
              jax.tree_util.tree_map(lambda x: None, state.params),
              0,
              None,
              None,
              jax.tree_util.tree_map(lambda x: None, predict_config),
              None,
              None,
          ),
      ),
      in_shardings=(
          data_sharding,
          state_mesh_annotations.params,
          data_sharding,
      ),  # type: ignore
      out_shardings=data_sharding,  # type: ignore
      static_argnums=(3, 4, 5, 6, 7),
  )

  # Main Train Loop
  # ---------------------------------------------------------------------------

  # We init the first set of dropout PRNG keys, but update it afterwards inside
  # the main pmap"d training update for performance.
  dropout_rngs = rng

  logging.info("Starting training loop.")
  hooks = []
  report_progress = periodic_actions.ReportProgress(
      num_train_steps=config.num_train_steps, writer=writer
  )
  if jax.process_index() == 0:
    hooks += [
        report_progress,
        periodic_actions.Profile(logdir=workdir, num_profile_steps=5),
    ]
  train_metrics = []
  with metric_writers.ensure_flushes(writer):
    for step in range(start_step, config.num_train_steps):
      is_last_step = step == config.num_train_steps - 1

      # Shard data to devices and do a training step.
      with jax.profiler.StepTraceAnnotation("train", step_num=step):
        batch = next(train_iter)
        batch = jax.tree_util.tree_map(lambda x: jnp.array(x), batch)
        state, metrics = jit_train_step(
            state, batch, train_config, learning_rate_fn, 0.0, dropout_rngs
        )
        train_metrics.append(metrics)

      # Quick indication that training is happening.
      logging.log_first_n(logging.INFO, "Finished training step %d.", 5, step)
      for h in hooks:
        h(step)

      # Periodic metric handling.
      if step % config.eval_every_steps == 0 or is_last_step:
        with report_progress.timed("training_metrics"):
          logging.info("Gathering training metrics.")
          train_metrics = common_utils.stack_forest(train_metrics)
          lr = train_metrics.pop("learning_rate").mean()
          metrics_sums = jax.tree_util.tree_map(jnp.sum, train_metrics)
          denominator = metrics_sums.pop("denominator")
          summary = jax.tree_util.tree_map(
              lambda x: x / denominator, metrics_sums
          )  # pylint: disable=cell-var-from-loop
          summary["learning_rate"] = lr
          summary["perplexity"] = jnp.clip(
              jnp.exp(summary["loss"]), max=1.0e4
          )
          summary = {"train_" + k: v for k, v in summary.items()}
          writer.write_scalars(step, summary)
          train_metrics = []

        with report_progress.timed("eval"):
          eval_results = evaluate(
              jit_eval_step=jit_eval_step,
              params=state.params,
              eval_ds=eval_ds,
              num_eval_steps=config.num_eval_steps,
              config=eval_config,
          )
          # (clipped) perplexity after averaging log-perplexitie
          eval_results["perplexity"] = jnp.clip(
              jnp.exp(eval_results["loss"]), max=1.0e4
          )
          writer.write_scalars(
              step, {"eval_" + k: v for k, v in eval_results.items()}
          )

        with report_progress.timed("generate_text"):
          exemplars = generate_prediction(
              jit_pred_step=jit_pred_step,
              params=state.params,
              tokenized_prompts=tokenized_prompts,
              eos_id=eos_id,
              inference_rng=inference_rng,
              decode_tokens=decode_tokens,
              config=config,
              predict_config=predict_config,
          )
          writer.write_texts(step, {"samples": exemplars})

      # Save a checkpoint on one host after every checkpoint_freq steps.
      save_checkpoint = (
          step % config.checkpoint_every_steps == 0 or is_last_step
      )
      if config.save_checkpoints and save_checkpoint:
        logging.info("Saving checkpoint step %d.", step)
        with report_progress.timed("checkpoint"):
          checkpoints.save_checkpoint_multiprocess(workdir, state, step)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilty functions used in data-parallel Flax examples.

This module is a historical grab-bag of utility functions primarily concerned
with helping write pmap-based data-parallel training loops.
"""

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax


def shard(xs):
  """Helper for pmap to shard a pytree of arrays by local_device_count.

  Args:
    xs: a pytree of arrays.
  Returns:
    A matching pytree with arrays' leading dimensions sharded by the
    local device count.
  """
  local_device_count = jax.local_device_count()
  return jax.tree_util.tree_map(
    lambda x: x.reshape((local_device_count, -1) + x.shape[1:]), xs
  )


def shard_prng_key(prng_key):
  """Helper to shard (aka split) a PRNGKey for use with pmap'd functions.

  PRNG keys can be used at train time to drive stochastic modules
  e.g. Dropout. We would like a different PRNG key for each local
  device so that we end up with different random numbers on each one,
  hence we split our PRNG key.

  Args:
    prng_key: JAX PRNGKey
  Returns:
    A new array of PRNGKeys with leading dimension equal to local device count.
  """
  return jax.random.split(prng_key, num=jax.local_device_count())


def stack_forest(forest):
  """Helper function to stack the leaves of a sequence of pytrees.

  Args:
    forest: a sequence of pytrees (e.g tuple or list) of matching structure
      whose leaves are arrays with individually matching shapes.
  Returns:
    A single pytree of the same structure whose leaves are individually
      stacked arrays.
  """
  stack_args = lambda *args: np.stack(args)
  return jax.tree_util.tree_map(stack_args, *forest)


def get_metrics(device_metrics):
  """Helper utility for pmap, gathering replicated timeseries metric data.

  Args:
   device_metrics: replicated, device-resident pytree of metric data,
     whose leaves are presumed to be a sequence of arrays recorded over time.
  Returns:
   A pytree of unreplicated, host-resident, stacked-over-time arrays useful for
   computing host-local statistics and logging.
  """
  # We select the first element of x in order to get a single copy of a
  # device-replicated metric.
  device_metrics = jax.tree_util.tree_map(lambda x: x[0], device_metrics)
  metrics_np = jax.device_get(device_metrics)
  return stack_forest(metrics_np)


def onehot(labels, num_classes, on_value=1.0, off_value=0.0):
  """Create a dense one-hot version of an indexed array.

  NB: consider using the more standard ``jax.nn.one_hot`` instead.

  Args:
    labels: an n-dim JAX array whose last dimension contains integer indices.
    num_classes: the maximum possible index.
    on_value: the "on" value for the one-hot array, defaults to 1.0.
    off_value: the "off" value for the one-hot array, defaults to 0.0.
  Returns:
    A (n+1)-dim array whose last dimension contains one-hot vectors of length
    num_classes.
  """
  x = labels[..., None] == jnp.arange(num_classes).reshape(
    (1,) * labels.ndim + (-1,)
  )
  x = lax.select(x, jnp.full(x.shape, on_value), jnp.full(x.shape, off_value))
  return x.astype(jnp.float32)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for sst2.models."""
from absl.testing import absltest
from absl.testing import parameterized
import models
import jax
import jax.test_util
import numpy as np


# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


class ModelTest(parameterized.TestCase):

  def test_embedder_returns_correct_output_shape(self):
    """Tests if the embedder returns the correct shape."""
    vocab_size = 5
    embedding_size = 3
    model = models.Embedder(
        vocab_size=vocab_size, embedding_size=embedding_size
    )
    rng = jax.random.key(0)
    token_ids = np.array([[2, 4, 3], [2, 6, 3]], dtype=np.int32)
    output, _ = model.init_with_output(rng, token_ids, deterministic=True)
    self.assertEqual((token_ids.shape) + (embedding_size,), output.shape)

  def test_lstm_returns_correct_output_shape(self):
    """Tests if the simple LSTM returns the correct shape."""
    batch_size = 2
    seq_len = 3
    embedding_size = 4
    hidden_size = 5
    model = models.SimpleLSTM(5)
    rng = jax.random.key(0)
    inputs = np.random.RandomState(0).normal(
        size=[batch_size, seq_len, embedding_size]
    )
    initial_state = model.initialize_carry(inputs[:, 0].shape)
    (_, output), _ = model.init_with_output(rng, initial_state, inputs)
    self.assertEqual((batch_size, seq_len, hidden_size), output.shape)

  def test_bilstm_returns_correct_output_shape(self):
    """Tests if the simple BiLSTM returns the correct shape."""
    batch_size = 2
    seq_len = 3
    embedding_size = 4
    hidden_size = 5
    model = models.SimpleBiLSTM(hidden_size=hidden_size)
    rng = jax.random.key(0)
    inputs = np.random.RandomState(0).normal(
        size=[batch_size, seq_len, embedding_size]
    )
    lengths = np.array([2, 3], dtype=np.int32)
    outputs, _ = model.init_with_output(rng, inputs, lengths)
    # We expect 2*hidden_size because we concatenate forward+backward LSTMs.
    self.assertEqual((batch_size, seq_len, 2 * hidden_size), outputs.shape)

  def test_text_classifier_returns_correct_output_shape(self):
    """Tests if a TextClassifier model returns the correct shape."""
    embedding_size = 3
    hidden_size = 7
    vocab_size = 5
    output_size = 3
    dropout_rate = 0.1
    word_dropout_rate = 0.2
    unk_idx = 1

    model = models.TextClassifier(
        embedding_size=embedding_size,
        hidden_size=hidden_size,
        vocab_size=vocab_size,
        output_size=output_size,
        dropout_rate=dropout_rate,
        word_dropout_rate=word_dropout_rate,
        unk_idx=unk_idx,
        deterministic=True,
    )

    rng = jax.random.key(0)
    token_ids = np.array([[2, 4, 3], [2, 6, 3]], dtype=np.int32)
    lengths = np.array([2, 3], dtype=np.int32)
    output, _ = model.init_with_output(rng, token_ids, lengths)
    batch_size = token_ids.shape[0]
    self.assertEqual((batch_size, output_size), output.shape)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Language Modeling example.

This script trains a Transformer on a LM1B dataset.
"""

# pytype: disable=wrong-arg-count
# pytype: disable=attribute-error

import collections
import dataclasses
import os

import input_pipeline
import jax
import jax.numpy as jnp
import models
import numpy as np
import optax
import temperature_sampler
import tensorflow as tf
import utils
from absl import logging
from clu import metric_writers, periodic_actions
from configs import default
from jax import random
from jax.sharding import Mesh, NamedSharding
from jax.sharding import PartitionSpec as P
from utils import HasCache, TrainState

from flax import linen as nn
from flax import nnx
from flax.training import checkpoints, common_utils


def rsqrt_schedule(
  init_value: float,
  shift: int = 0,
):
  """Applies a reverse square-root schedule.

  The reverse square root schedule is simply `lr = init_value / sqrt(step)`.

  Args:
    init_value: Base learning rate (before applying the rsqrt schedule).
    shift: How many steps the rsqrt should be shifted. Shifting the rsqrt
      schedule makes it less steep in the beginning (close to 0).

  Returns:
    A schedule that applies the reverse square root.
  """

  def schedule(count):
    return init_value * (count + shift) ** -0.5 * shift**0.5

  return schedule


def create_learning_rate_schedule(learning_rate: float, warmup_steps: int):
  """Creates a rsqrt schedule with linear warmup."""
  return optax.join_schedules(
    [
      optax.linear_schedule(
        init_value=0,
        end_value=learning_rate,
        transition_steps=warmup_steps,
      ),
      rsqrt_schedule(init_value=learning_rate, shift=warmup_steps),
    ],
    boundaries=[warmup_steps],
  )


def compute_weighted_cross_entropy(
  logits, targets, weights=None, label_smoothing=0.0
):
  """Compute weighted cross entropy and entropy for log probs and targets.

  Args:
   logits: [batch, length, num_classes] float array.
   targets: categorical targets [batch, length] int array.
   weights: None or array of shape [batch, length].
   label_smoothing: label smoothing constant, used to determine the on and off
     values.

  Returns:
    Tuple of scalar loss and batch normalizing factor.
  """
  if logits.ndim != targets.ndim + 1:
    raise ValueError(
      'Incorrect shapes. Got shape %s logits and %s targets'
      % (str(logits.shape), str(targets.shape))
    )
  vocab_size = logits.shape[-1]
  confidence = 1.0 - label_smoothing
  low_confidence = (1.0 - confidence) / (vocab_size - 1)
  normalizing_constant = -(
    confidence * jnp.log(confidence)
    + (vocab_size - 1) * low_confidence * jnp.log(low_confidence + 1e-20)
  )
  soft_targets = common_utils.onehot(
    targets, vocab_size, on_value=confidence, off_value=low_confidence
  )

  loss = -jnp.sum(soft_targets * nn.log_softmax(logits), axis=-1)
  loss = loss - normalizing_constant

  normalizing_factor = np.prod(targets.shape)
  if weights is not None:
    loss = loss * weights
    normalizing_factor = weights.sum()

  return loss.sum(), normalizing_factor


def compute_weighted_accuracy(logits, targets, weights=None):
  """Compute weighted accuracy for log probs and targets.

  Args:
   logits: [batch, length, num_classes] float array.
   targets: categorical targets [batch, length] int array.
   weights: None or array of shape [batch, length]

  Returns:
    Tuple of scalar loss and batch normalizing factor.
  """
  if logits.ndim != targets.ndim + 1:
    raise ValueError(
      'Incorrect shapes. Got shape %s logits and %s targets'
      % (str(logits.shape), str(targets.shape))
    )
  loss = jnp.equal(jnp.argmax(logits, axis=-1), targets)
  normalizing_factor = np.prod(logits.shape[:-1])
  if weights is not None:
    loss = loss * weights
    normalizing_factor = weights.sum()

  return loss.sum(), normalizing_factor


def compute_metrics(logits, labels, weights, label_smoothing=0.0):
  """Compute summary metrics."""
  loss, weight_sum = compute_weighted_cross_entropy(
    logits, labels, weights, label_smoothing
  )
  acc, _ = compute_weighted_accuracy(logits, labels, weights)
  metrics = {
    'loss': loss,
    'accuracy': acc,
    'denominator': weight_sum,
  }
  return metrics


# Primary training / eval / decode step functions.
# -----------------------------------------------------------------------------


def train_step(
  state: TrainState,
  batch,
  learning_rate_fn,
  label_smoothing=0.0,
  dropout_rng=None,
):
  """Perform a single training step."""
  # X_position and X_segmentation are needed only when using "packed examples"
  # where multiple sequences are packed into the same example with this
  # metadata.
  # if such features are not present they are ignored and the example is treated
  # like a normal, unpacked sequence example.
  train_keys = ['inputs', 'inputs_position', 'inputs_segmentation']
  (inputs, inputs_positions, inputs_segmentation) = (
    batch.get(k, None) for k in train_keys
  )

  weights = jnp.where(inputs > 0, 1, 0).astype(jnp.float32)

  dropout_rng = jax.random.fold_in(dropout_rng, state.step)

  def loss_fn(params):
    """loss function used for training."""
    module = nnx.merge(state.graphdef, params)
    module.set_attributes(deterministic=False, decode=False)
    logits = module(
      inputs,
      inputs_positions=inputs_positions,
      inputs_segmentation=inputs_segmentation,
      rngs=nnx.Rngs(dropout=dropout_rng),
    )

    loss, weight_sum = compute_weighted_cross_entropy(
      logits, inputs, weights, label_smoothing
    )
    mean_loss = loss / weight_sum
    return mean_loss, logits

  step = state.step
  lr = learning_rate_fn(step)
  grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  (_, logits), grads = grad_fn(state.params)
  new_state = state.apply_gradients(grads=grads)
  metrics = compute_metrics(logits, inputs, weights)
  metrics['learning_rate'] = lr

  return new_state, metrics


def eval_step(
  params: nnx.State,
  batch,
  graphdef: nnx.GraphDef[models.TransformerLM],
  label_smoothing=0.0,
):
  """Calculate evaluation metrics on a batch."""
  inputs = batch['inputs']
  weights = jnp.where(inputs > 0, 1.0, 0.0)
  module = nnx.merge(graphdef, params)
  module.set_attributes(deterministic=True, decode=False)
  logits = module(inputs)

  return compute_metrics(logits, inputs, weights, label_smoothing)


def predict_step(
  inputs,
  params: nnx.State,
  rngkey: jax.Array,
  graphdef: nnx.GraphDef[models.TransformerLM],
  eos_id: int,
  max_decode_len: int,
  config: models.TransformerConfig,
  temperature: float,
  top_k: int,
):
  """Predict language model on a batch."""
  module = nnx.merge(graphdef, params)

  # TODO(cgarciae): check how pytorch does this.
  for _path, m in module.iter_modules():
    if isinstance(m, HasCache):
      input_shape = (inputs.shape[0], max_decode_len, config.emb_dim)
      m.init_cache(input_shape, dtype=config.dtype)

  graphdef, params, cache = nnx.split(module, nnx.Param, nnx.Cache)

  def tokens_ids_to_logits(flat_ids, cache: nnx.State):
    """Token slice to logits from decoder model."""
    # --> [batch * beam, 1, vocab]
    module = nnx.merge(graphdef, params, cache)
    module.set_attributes(deterministic=True, decode=True)
    logits = module(flat_ids)
    cache = nnx.state(module, nnx.Cache)
    # Remove singleton sequence-length dimension:
    # [batch, 1, vocab] --> [batch, vocab]
    logits = logits.squeeze(axis=1)
    return logits, cache

  # Using the above-defined single-step decoder function, run a
  # beam search over possible sequences given input encoding.
  seqs = temperature_sampler.temperature_sample(
    inputs,
    cache,
    tokens_ids_to_logits,
    rngkey,
    temperature=temperature,
    topk=top_k,
    eos_token=eos_id,
  )

  return seqs


# Utils for prediction
# -----------------------------------------------------------------------------


def pad_examples(x, desired_batch_size):
  """Expand batch to desired size by repeating last slice."""
  batch_pad = desired_batch_size - x.shape[0]
  return np.concatenate([x, np.tile(x[-1], (batch_pad, 1))], axis=0)


def per_host_sum_pmap(in_tree):
  """Execute psum on in_tree"s leaves over one device per host."""
  host2devices = collections.defaultdict(list)
  for d in jax.devices():
    host2devices[d.process_index].append(d)
  devices = [host2devices[k][0] for k in host2devices]
  host_psum = jax.pmap(lambda x: jax.lax.psum(x, 'i'), 'i', devices=devices)

  def pre_pmap(xs):
    return jax.tree_util.tree_map(
      lambda x: jnp.broadcast_to(x, (1,) + x.shape), xs
    )

  def post_pmap(xs):
    return jax.tree_util.tree_map(lambda x: x[0], xs)

  return post_pmap(host_psum(pre_pmap(in_tree)))


def tohost(x):
  """Collect batches from all devices to host and flatten batch dimensions."""
  n_device, n_batch, *remaining_dims = x.shape
  return np.array(x).reshape((n_device * n_batch,) + tuple(remaining_dims))


def evaluate(
  *,
  jit_eval_step,
  state: TrainState,
  eval_ds: tf.data.Dataset,
  num_eval_steps: int,
):
  """Evaluate the target an return a dictionary with the metrics."""
  logging.info('Gathering evaluation metrics.')
  eval_metrics = []
  eval_iter = iter(eval_ds)  # pytype: disable=wrong-arg-types
  for _, eval_batch in zip(range(num_eval_steps), eval_iter):
    eval_batch = jax.tree_util.tree_map(lambda x: x._numpy(), eval_batch)  # pylint: disable=protected-access
    metrics = jit_eval_step(state.params, eval_batch, state.graphdef)
    eval_metrics.append(metrics)
  eval_metrics = common_utils.stack_forest(eval_metrics)
  eval_metrics_sums = jax.tree_util.tree_map(jnp.sum, eval_metrics)
  eval_denominator = eval_metrics_sums.pop('denominator')
  eval_summary = jax.tree_util.tree_map(
    lambda x: x / eval_denominator,  # pylint: disable=cell-var-from-loop
    eval_metrics_sums,
  )
  return eval_summary


def generate_prediction(
  *,
  jit_pred_step,
  graphdef: nnx.GraphDef[models.TransformerLM],
  params: nnx.State,
  tokenized_prompts,
  eos_id,
  inference_rng,
  decode_tokens,
  config: default.Config,
  model_config: models.TransformerConfig,
):
  """Generate text from the prompt."""
  n_devices = jax.local_device_count()

  logging.info('Generating text.')
  predictions = []
  # Use batch of prompts provided by user.
  for pred_batch in jnp.array_split(
    tokenized_prompts, int(np.ceil(len(tokenized_prompts) / n_devices))
  ):
    cur_pred_batch_size = pred_batch.shape[0]
    if cur_pred_batch_size % n_devices:
      padded_size = int(np.ceil(cur_pred_batch_size / n_devices) * n_devices)
      pred_batch = jax.tree_util.tree_map(
        lambda x: pad_examples(x, padded_size), pred_batch
      )  # pylint: disable=cell-var-from-loop
    pred_batch = common_utils.shard(pred_batch)
    inference_rng, sub_rng = random.split(inference_rng)
    inference_rngs = random.split(sub_rng, n_devices)

    predicted = jit_pred_step(
      pred_batch,
      params,
      inference_rngs,
      graphdef,
      eos_id,
      config.max_predict_length,
      model_config,
      config.sampling_temperature,
      config.sampling_top_k,
    )
    predicted = tohost(predicted)
    # Iterate through non-padding examples of batch.
    for s in predicted[:cur_pred_batch_size]:
      prediction = decode_tokens(s)
      logging.info('Sample: %s', str(prediction))
      predictions.append(prediction)

    # Save generated texts for tensorboard.
    exemplars = ''
    for prediction in predictions:
      exemplars += f'{prediction}\n\n'
  return exemplars


def train_and_evaluate(config: default.Config, workdir: str):
  """Runs a training and evaluation loop.

  Args:
    config: Configuration to use.
    workdir: Working directory for checkpoints and TF summaries. If this
      contains checkpoint training will be resumed from the latest checkpoint.
  """
  tf.io.gfile.makedirs(workdir)

  vocab_path = config.vocab_path
  if vocab_path is None:
    vocab_path = os.path.join(workdir, 'sentencepiece_model')
    config.vocab_path = vocab_path
  tf.io.gfile.makedirs(os.path.split(vocab_path)[0])

  # Load Dataset
  # ---------------------------------------------------------------------------
  logging.info('Initializing dataset.')
  train_ds, eval_ds, _, encoder = input_pipeline.get_datasets(
    n_devices=jax.local_device_count(), config=config, vocab_path=vocab_path
  )

  train_iter = iter(train_ds)
  vocab_size = int(encoder.vocab_size())
  eos_id = temperature_sampler.EOS_ID  # Default Sentencepiece EOS token.

  def decode_tokens(toks):
    valid_toks = toks[: np.argmax(toks == eos_id) + 1].astype(np.int32)
    return encoder.detokenize(valid_toks).numpy().decode('utf-8')

  def encode_strings(strs, max_len):
    tokenized_batch = np.zeros((len(strs), max_len), np.int32)
    for i, s in enumerate(strs):
      toks = encoder.tokenize(s).numpy()
      # Remove EOS token in prompt.
      tokenized_batch[i, : toks.shape[0] - 1] = toks[:-1]
    return tokenized_batch

  tokenized_prompts = encode_strings(
    [config.prompts], config.max_predict_length
  )

  logging.info('Initializing model, optimizer, and step functions.')
  # Build Model and Optimizer
  # ---------------------------------------------------------------------------
  model_config = models.TransformerConfig(
    vocab_size=vocab_size,
    output_vocab_size=vocab_size,
    logits_via_embedding=config.logits_via_embedding,
    dtype=jnp.bfloat16 if config.use_bfloat16 else jnp.float32,
    emb_dim=config.emb_dim,
    num_heads=config.num_heads,
    num_layers=config.num_layers,
    qkv_dim=config.qkv_dim,
    mlp_dim=config.mlp_dim,
    max_len=max(config.max_target_length, config.max_eval_target_length),
    dropout_rate=config.dropout_rate,
    attention_dropout_rate=config.attention_dropout_rate,
    kernel_init=nn.initializers.xavier_uniform(),
    bias_init=nn.initializers.normal(stddev=1e-6),
    axis_rules=config.axis_rules,
  )

  # Mesh definition
  devices_array = utils.create_device_mesh(config)
  mesh = Mesh(devices_array, config.mesh_axes)

  # print(mesh.shape)
  # exit()

  start_step = 0
  rng = jax.random.PRNGKey(config.seed)
  rng, init_rng = jax.random.split(rng)
  rng, inference_rng = random.split(rng)

  def constructor(config: models.TransformerConfig, key: jax.Array):
    return models.TransformerLM(config, rngs=nnx.Rngs(params=key))

  learning_rate_fn = create_learning_rate_schedule(
    learning_rate=config.learning_rate, warmup_steps=config.warmup_steps
  )

  optimizer = optax.adamw(
    learning_rate_fn,
    b1=0.9,
    b2=0.98,
    eps=1e-9,
    weight_decay=config.weight_decay,
  )

  state, state_sharding = utils.setup_initial_state(
    constructor, optimizer, model_config, init_rng, mesh
  )
  data_sharding = NamedSharding(mesh, P(config.data_sharding))

  if config.restore_checkpoints:
    # Restore unreplicated optimizer + model state from last checkpoint.
    state = checkpoints.restore_checkpoint(workdir, state)
    # Grab last step.
    start_step = int(state.step)

  writer = metric_writers.create_default_writer(
    workdir, just_logging=jax.process_index() > 0
  )
  if start_step == 0:
    writer.write_hparams(dataclasses.asdict(config))

  # compile multidevice versions of train/eval/predict step fn.
  jit_train_step = jax.jit(
    train_step,
    in_shardings=(
      state_sharding,
      data_sharding,
      None,
    ),  # type: ignore
    out_shardings=(state_sharding, None),  # type: ignore
    static_argnums=(2, 3),
    donate_argnums=0,
  )

  jit_eval_step = jax.jit(
    eval_step,
    in_shardings=(
      state_sharding.params,
      data_sharding,
    ),  # type: ignore
    out_shardings=None,  # type: ignore
    static_argnums=(2, 3),
  )

  # Since the inputs and rngkey args for predict_step will be batched,
  # we must vmap them, otherwise the global arrays will be seen in each device
  jit_pred_step = jax.jit(
    jax.vmap(
      predict_step,
      in_axes=(
        0,
        jax.tree_util.tree_map(lambda x: None, state.params),
        0,
        None,
        None,
        None,
        None,
        None,
        None,
      ),
    ),
    in_shardings=(
      data_sharding,
      state_sharding.params,
      data_sharding,
    ),  # type: ignore
    out_shardings=data_sharding,  # type: ignore
    static_argnums=tuple(range(3, 9)),
  )

  # Main Train Loop
  # ---------------------------------------------------------------------------

  # We init the first set of dropout PRNG keys, but update it afterwards inside
  # the main pmap"d training update for performance.
  dropout_rngs = rng

  logging.info('Starting training loop.')
  hooks = []
  report_progress = periodic_actions.ReportProgress(
    num_train_steps=config.num_train_steps, writer=writer
  )
  if jax.process_index() == 0:
    hooks += [
      report_progress,
      periodic_actions.Profile(logdir=workdir, num_profile_steps=5),
    ]
  train_metrics = []
  with metric_writers.ensure_flushes(writer):
    for step in range(start_step, config.num_train_steps):
      is_last_step = step == config.num_train_steps - 1

      # Shard data to devices and do a training step.
      with jax.profiler.StepTraceAnnotation('train', step_num=step):
        batch = next(train_iter)
        batch = jax.tree_util.tree_map(lambda x: jnp.asarray(x), batch)
        state, metrics = jit_train_step(
          state, batch, learning_rate_fn, 0.0, dropout_rngs
        )
        train_metrics.append(metrics)

      # Quick indication that training is happening.
      logging.log_first_n(logging.INFO, 'Finished training step %d.', 5, step)
      for h in hooks:
        h(step)

      # Periodic metric handling.
      if step % config.eval_every_steps == 0 or is_last_step:
        with report_progress.timed('training_metrics'):
          logging.info('Gathering training metrics.')
          train_metrics = common_utils.stack_forest(train_metrics)
          lr = train_metrics.pop('learning_rate').mean()
          metrics_sums = jax.tree_util.tree_map(jnp.sum, train_metrics)
          denominator = metrics_sums.pop('denominator')
          summary = jax.tree_util.tree_map(
            lambda x: x / denominator, metrics_sums
          )  # pylint: disable=cell-var-from-loop
          summary['learning_rate'] = lr
          summary['perplexity'] = jnp.clip(jnp.exp(summary['loss']), max=1.0e4)
          summary = {'train_' + k: v for k, v in summary.items()}
          writer.write_scalars(step, summary)
          train_metrics = []

        with report_progress.timed('eval'):
          eval_results = evaluate(
            jit_eval_step=jit_eval_step,
            state=state,
            eval_ds=eval_ds,
            num_eval_steps=config.num_eval_steps,
          )
          # (clipped) perplexity after averaging log-perplexitie
          eval_results['perplexity'] = jnp.clip(
            jnp.exp(eval_results['loss']), max=1.0e4
          )
          writer.write_scalars(
            step, {'eval_' + k: v for k, v in eval_results.items()}
          )

        with report_progress.timed('generate_text'):
          exemplars = generate_prediction(
            jit_pred_step=jit_pred_step,
            graphdef=state.graphdef,
            params=state.params,
            tokenized_prompts=tokenized_prompts,
            eos_id=eos_id,
            inference_rng=inference_rng,
            decode_tokens=decode_tokens,
            config=config,
            model_config=model_config,
          )
          writer.write_texts(step, {'samples': exemplars})

      # Save a checkpoint on one host after every checkpoint_freq steps.
      save_checkpoint = (
        step % config.checkpoint_every_steps == 0 or is_last_step
      )
      if config.save_checkpoints and save_checkpoint:
        logging.info('Saving checkpoint step %d.', step)
        with report_progress.timed('checkpoint'):
          checkpoints.save_checkpoint_multiprocess(workdir, state, step)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Sequence

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax.core import Array, Scope, init, lift, nn, unfreeze


def mlp_vmap(
  scope: Scope,
  x: Array,
  sizes: Sequence[int] = (8, 1),
  act_fn: Callable[[Array], Array] = nn.relu,
  share_params: bool = False,
):
  if share_params:
    dense_vmap = lift.vmap(
      nn.dense,
      in_axes=(0, None),
      variable_axes={'params': None},
      split_rngs={'params': False},
    )
  else:
    dense_vmap = lift.vmap(
      nn.dense,
      in_axes=(0, None),
      variable_axes={'params': 0},
      split_rngs={'params': True},
    )

  # hidden layers
  for size in sizes[:-1]:
    x = scope.child(dense_vmap, prefix='hidden_')(x, size)
    x = act_fn(x)

  # output layer
  return scope.child(dense_vmap, 'out')(x, sizes[-1])


class VMapTest(absltest.TestCase):
  def test_vmap_shared(self):
    x = random.normal(random.key(0), (1, 4))
    x = jnp.concatenate([x, x], 0)

    y, variables = init(mlp_vmap)(random.key(1), x, share_params=True)

    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(
      param_shapes,
      {
        'hidden_0': {'kernel': (4, 8), 'bias': (8,)},
        'out': {'kernel': (8, 1), 'bias': (1,)},
      },
    )
    self.assertEqual(y.shape, (2, 1))
    self.assertTrue(jnp.allclose(y[0], y[1]))

  def test_vmap_unshared(self):
    x = random.normal(random.key(0), (1, 4))
    x = jnp.concatenate([x, x], 0)

    y, variables = init(mlp_vmap)(random.key(1), x, share_params=False)

    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(
      param_shapes,
      {
        'hidden_0': {'kernel': (2, 4, 8), 'bias': (2, 8)},
        'out': {'kernel': (2, 8, 1), 'bias': (2, 1)},
      },
    )
    self.assertEqual(y.shape, (2, 1))
    self.assertFalse(jnp.allclose(y[0], y[1]))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Transformer sub-modules."""

from __future__ import annotations

import enum
from typing import Any, Sequence, Union

from flax import nnx
import flax.linen as nn
import layers
import positional_embeddings
import jax
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike  # pylint: disable=g-importing-member,g-multiple-import

LayerCache = dict[str, Array]
Shape = Sequence[Union[int, Any]]

K_MASK = -2.3819763e38  # Set to a large negative number.


class AttentionType(enum.Enum):
  GLOBAL = 1
  LOCAL_SLIDING = 2


class Embedder(nnx.Module):
  """Embedder module."""

  def __init__(
      self,
      vocab_size: int,
      embed_dim: int,
      *,
      rngs: nnx.Rngs,
  ):
    self.input_embedding = nnx.Param(
        nn.initializers.normal()(rngs.params(), (vocab_size, embed_dim))
    )

  def encode(self, x: ArrayLike) -> Array:
    x = self.input_embedding[(x,)]
    x *= jnp.sqrt(x.shape[-1]).astype(x.dtype)
    return x

  def decode(self, x: ArrayLike) -> Array:
    return jnp.dot(x, self.input_embedding.value.T)

  @property
  def embed_dim(self):
    return self.input_embedding.value.shape[1]

  @property
  def num_embed(self):
    return self.input_embedding.value.shape[0]


class Attention(nnx.Module):
  """Attention module."""

  def __init__(
      self,
      num_heads: int,
      num_kv_heads: int,
      features: int,
      head_dim: int,
      attn_type: AttentionType,
      *,
      rngs: nnx.Rngs,
      attn_logits_soft_cap: Union[float, None] = None,
      sliding_window_size: Union[int, None] = None,
  ):
    if attn_type == AttentionType.LOCAL_SLIDING and sliding_window_size is None:
      raise ValueError(
          '`sliding_window_size` must be set if `attn_type` is Local Sliding.'
      )

    self.attn_type = attn_type
    self.sliding_window_size = sliding_window_size
    self.attn_logits_soft_cap = attn_logits_soft_cap
    self.attn_vec_einsum = layers.Einsum(
        shape=(num_heads, head_dim, features),
        rngs=rngs,
    )

    if num_heads == num_kv_heads:
      self.qkv_einsum = layers.Einsum(
          shape=(3, num_heads, features, head_dim),
          rngs=rngs,
      )
    else:
      self.q_einsum = layers.Einsum(
          shape=(num_heads, features, head_dim),
          rngs=rngs,
      )
      self.kv_einsum = layers.Einsum(
          shape=(2, num_kv_heads, features, head_dim),
          rngs=rngs,
      )

  def __call__(
      self,
      x: Array,
      segment_pos: Array,
      cache: Union[LayerCache, None],
      attn_mask: Array,
  ) -> tuple[Union[LayerCache, None], Array]:
    seq_len = x.shape[1]

    if self.use_qkv_einsum:
      query_proj, key_proj, value_proj = self.qkv_einsum('BTD,SNDH->SBTNH', x)
    else:
      query_proj = self.q_einsum('BTD,NDH->BTNH', x)
      key_proj, value_proj = self.kv_einsum('BSD,CKDH->CBSKH', x)

    query_proj = positional_embeddings.apply_rope(
        query_proj,
        segment_pos,
        head_dim=self.head_dim,
    )
    query_scaled = query_proj * self.head_dim**-0.5
    key_proj = positional_embeddings.apply_rope(
        key_proj,
        segment_pos,
        head_dim=self.head_dim,
    )

    # Cache is left aligned.
    if cache is not None:
      end_index = cache['end_index'][0]
      slice_indices = (0, end_index % cache['v'].shape[1], 0, 0)
      value_proj = jax.lax.dynamic_update_slice(
          cache['v'],
          value_proj,
          slice_indices,
      )
      key_proj = jax.lax.dynamic_update_slice(
          cache['k'], key_proj, slice_indices
      )

    logits = jnp.einsum('BTNH,BSNH->BTNS', query_scaled, key_proj)

    if self.attn_logits_soft_cap is not None:
      logits = jnp.tanh(logits / self.attn_logits_soft_cap)
      logits = logits * self.attn_logits_soft_cap
    if self.attn_type == AttentionType.LOCAL_SLIDING:
      if self.sliding_window_size is None:
        raise ValueError(
            'sliding_window_size must be set if attn_type is Local Sliding.'
        )

      all_ones = jnp.ones_like(attn_mask)
      sliding_mask = jnp.triu(
          all_ones, -1 * self.sliding_window_size + 1
      ) * jnp.tril(all_ones, self.sliding_window_size - 1)
      attn_mask = sliding_mask * attn_mask

    padded_logits = jnp.where((jnp.expand_dims(attn_mask, -2)), logits, K_MASK)
    probs = jax.nn.softmax(padded_logits, axis=-1).astype(key_proj.dtype)
    encoded = jnp.einsum('BTNS,BSNH->BTNH', probs, value_proj)
    attn_output = self.attn_vec_einsum('BTNH,NHD->BTD', encoded)

    if cache is not None:
      new_cache = {
          'v': value_proj,
          'k': key_proj,
          'end_index': cache['end_index'] + seq_len,
      }
    else:
      new_cache = None

    return new_cache, attn_output

  @property
  def head_dim(self):
    return self.attn_vec_einsum.shape[1]

  @property
  def num_heads(self):
    return (
        self.qkv_einsum.shape[1]
        if self.use_qkv_einsum
        else self.q_einsum.shape[0]
    )

  @property
  def num_kv_heads(self):
    return (
        self.qkv_einsum.shape[1]
        if self.use_qkv_einsum
        else self.kv_einsum.shape[1]
    )

  @property
  def use_qkv_einsum(self):
    return hasattr(self, 'qkv_einsum') and self.qkv_einsum is not None

  def init_cache(
      self,
      cache_size: int,
      batch_size: int,
      dtype: jnp.dtype = jnp.bfloat16,
  ) -> LayerCache:
    return {
        'v': jnp.zeros(
            (batch_size, cache_size, self.num_kv_heads, self.head_dim),
            dtype=dtype,
        ),
        'k': jnp.zeros(
            (batch_size, cache_size, self.num_kv_heads, self.head_dim),
            dtype=dtype,
        ),
        'end_index': jnp.zeros((batch_size,), dtype=jnp.int32),
    }


class FeedForward(nnx.Module):
  """Feed forward module."""

  def __init__(
      self,
      features: int,
      hidden_dim: int,
      *,
      rngs: nnx.Rngs,
  ):
    self.gating_einsum = nnx.Param(
        nn.initializers.zeros_init()(
            rngs.params(),
            ((2, features, hidden_dim)),
        )
    )
    self.linear = nnx.Param(
        nn.initializers.zeros_init()(
            rngs.params(),
            (hidden_dim, features),
        )
    )

  def __call__(self, x: ArrayLike) -> Array:
    ff_gate = jnp.dot(x, self.gating_einsum.value[0])
    gate_value = nnx.gelu(ff_gate)

    ff1 = jnp.dot(x, self.gating_einsum.value[1])
    activations = gate_value * ff1

    outputs = jnp.dot(activations, self.linear.value)
    return outputs


class Block(nnx.Module):
  """Transformer block."""

  def __init__(
      self,
      num_heads: int,
      num_kv_heads: int,
      embed_dim: int,
      head_dim: int,
      hidden_dim: int,
      use_post_attn_norm: bool,
      use_post_ffw_norm: bool,
      attn_type: AttentionType,
      *,
      rngs: nnx.Rngs,
      attn_logits_soft_cap: Union[float, None] = None,
      sliding_window_size: Union[int, None] = None,
  ):
    self.pre_attention_norm = layers.RMSNorm(embed_dim, rngs=rngs)
    self.attn = Attention(
        num_heads=num_heads,
        num_kv_heads=num_kv_heads,
        features=embed_dim,
        head_dim=head_dim,
        attn_type=attn_type,
        attn_logits_soft_cap=attn_logits_soft_cap,
        sliding_window_size=sliding_window_size,
        rngs=rngs,
    )
    if use_post_attn_norm:
      self.post_attn_norm = layers.RMSNorm(embed_dim, rngs=rngs)

    self.pre_ffw_norm = layers.RMSNorm(embed_dim, rngs=rngs)
    self.mlp = FeedForward(
        features=embed_dim,
        hidden_dim=hidden_dim,
        rngs=rngs,
    )
    if use_post_ffw_norm:
      self.post_ffw_norm = layers.RMSNorm(embed_dim, rngs=rngs)

  def __call__(
      self,
      x: jax.Array,
      segment_pos: jax.Array,
      cache: Union[LayerCache, None],
      attn_mask: jax.Array,
  ) -> tuple[Union[LayerCache, None], jax.Array]:
    inputs_normalized = self.pre_attention_norm(x)
    cache, attn_output = self.attn(
        inputs_normalized,
        segment_pos,
        cache,
        attn_mask,
    )
    attn_output += x
    residual = attn_output
    attn_output = self.pre_ffw_norm(attn_output)

    if self.use_post_attn_norm:
      attn_output = self.post_attn_norm(attn_output)

    outputs = self.mlp(attn_output)
    if self.use_post_ffw_norm:
      outputs = self.post_ffw_norm(outputs)
    outputs = residual + outputs
    return cache, outputs

  @property
  def use_post_attn_norm(self):
    return hasattr(self, 'post_attn_norm') and self.post_attn_norm is not None

  @property
  def use_post_ffw_norm(self):
    return hasattr(self, 'post_ffw_norm') and self.post_ffw_norm is not None

  def init_cache(
      self,
      cache_size: int,
      batch_size: int,
      dtype: jnp.dtype = jnp.bfloat16,
  ) -> LayerCache:
    return self.attn.init_cache(
        cache_size=cache_size,
        batch_size=batch_size,
        dtype=dtype,
    )
